#!/usr/local/bin/python3.13
# License: GPL v3 Copyright: 2021, Kovid Goyal <kovid at kovidgoyal.net>


import ast

from calibre.srv.tests.base import SimpleTest
from calibre_extensions.fast_css_transform import parse_css_number, transform_properties


class TestTransform(SimpleTest):

    def test_counting_chars_in_elems(self):
        from lxml import etree

        from calibre.ebooks.oeb.polish.parsing import parse
        from calibre.srv.render_book import get_length
        def t(html, expected):
            root = parse(html, force_html5_parse=True)
            self.assertEqual(expected, get_length(root), etree.tostring(root, encoding=str))
        t('<p>abc<span>def</span>x yz<svg>howdy', 1014)
        t('<p>abc<span>def</span>x yz', 9)
        t('<p>abc<span>def</span><script>x yz', 6)
        t('<p>abc<span>def</span><style>x yz', 6)
        t('<p>abc<span>def</span>x yz<img>howdy', 1014)
        t('<p>abc<span>def</span>x yz<img><!-- howdy -->howdy', 1014)

    def test_number_parsing(self):
        for x in '.314 -.314 0.314 0 2 +2 -1 1e2 -3.14E+2 2e-2'.split():
            self.ae(parse_css_number(x), ast.literal_eval(x))
        self.ae(parse_css_number('2em'), 2)
        self.ae(parse_css_number('.3em'), 0.3)
        self.ae(parse_css_number('3x3'), 3)

    def test_basic_css_transforms(self):
        def d(src, expected, is_declaration=True, url_callback=None):
            self.ae(transform_properties(src, is_declaration=is_declaration, url_callback=url_callback), expected)

        def upper_case(val):
            return val.upper()

        def u(src, expected, is_declaration=True, url_callback=upper_case):
            return d(src, expected, url_callback=url_callback, is_declaration=is_declaration)

        def s(src, expected, url_callback=upper_case):
            return d(src, expected, url_callback=url_callback, is_declaration=False)

        s('.c { page-break-after: 1 always }', '.c { break-after: 1 always ; -webkit-column-break-after: 1 always }')
        s('.c { page-break-after: always\ncolor:red }', '.c { break-after: always; -webkit-column-break-after: always\ncolor:red }')
        s('.c { page-break-after: always\n}', '.c { break-after: always; -webkit-column-break-after: always\n}')
        s('.c { page-break-after: always;color:red }', '.c { break-after: always; -webkit-column-break-after: always;color:red }')
        s('.c { page-break-after: /**/always }', '.c { break-after: always ; -webkit-column-break-after: always }')
        s('.c { page-break-after: always !important }', '.c { break-after: always !important ; -webkit-column-break-after: always !important }')
        s('.c { page-break-after: always;}', '.c { break-after: always; -webkit-column-break-after: always;}')
        s('.c{x:url(y)}', '.c{x:url("Y")}')
        s('@im/* c */port "x.y";', '@import "X.Y";')
        s('@import url("narrow.css") supports(display: flex) handheld and (max-width: 400px);',
          '@import url("NARROW.CSS") supports(display: flex) handheld and (max-width: 400px);')
        s('@import url( x/*a*/.y);', '@import url("X.Y");')

        u('background: url(  te  st.gif  ) 12; src: url(x)', 'background: url("TE  ST.GIF") 12; src: url("X")')
        u('background: url(te/**/st.gif); xxx: url()', 'background: url("TEST.GIF"); xxx: url()')
        u(r'background: uRl(t\)e/st.gif)', 'background: url("T)E/ST.GIF")')
        u('a:url(  "( )" /**/ )', 'a:url("( )")')
        u('a:url(  "(/*)"  )', 'a:url(  "(/*)"  )', url_callback=lambda x: x)

        d('font-size: 197583965730245.28px', 'font-size: 1.2349e+13rem')
        d('font-size: 19.28px', 'font-size: 1.205rem')
        d('font-size:+19.28px', 'font-size:1.205rem')
        d('font-size: .28in', 'font-size: 1.68rem')
        d('font-size: +.28in', 'font-size: 1.68rem')
        d(r'f\ont-s\69z\65 : 16\px', 'font-size: 1rem')
        d('font -size: 16px', 'font -size: 16px')
        d('font-/* */size: 1/*x*/6/**/p/**/x !important', 'font-size: 1rem !important')
        d('fOnt-size :16px', 'fOnt-size :1rem')
        d('fönt-size :16px', 'fönt-size :16px')
        d('font-size:2%', 'font-size:2%')
        d('font-size: 72pt; margin: /*here*/ 20px; font-size: 2in', 'font-size: 6rem; margin: /*here*/ 20px; font-size: 12rem')
        d(r'''font: "some 'name" 32px''', 'font: "some \'name" 2rem')
        d(r'''font: 'some "name' 32px''', 'font: \'some "name\' 2rem')
        d(r'''font: 'some \n ame' 32px''', 'font: "some n ame" 2rem')
        d('''font: 'some \\\nname' 32px''', 'font: "some name" 2rem')
        d('font: sans-serif 16px/3', 'font: sans-serif 1rem/3')
        d('font: sans-serif small/17', 'font: sans-serif 0.8rem/17')

        d('-epub-writing-mode: a; -web/* */kit-writing-mode: b; writing-mode: c', 'writing-mode: a; writing-mode: b; writing-mode: c')
        d('xxx:yyy', 'xxx:yyy')
        d('page-break-before: always', 'break-before: always; -webkit-column-break-before: always')

        sheet = '''
@import "b/loc.test";
@media screen {
    font: 16px calc(20vw - 30rem);

    .cls {
        color: red;
        font-size: 16px;
        background: url("b/loc.test")
    }

    #moo.cat {
        x: url("b/loc.test")
    }

    @zoo {
        not(.woo) and why {
            font: 16px "something something" 16;
            page-break-before: avoid
        }
    }
}
.why { font: 16px}
'''
        s(sheet, sheet.replace('16px', '1rem').replace('b/loc.test', 'B/LOC.TEST').replace('page-', 'break-before: avoid; -webkit-column-'))



#!/usr/local/bin/python3.13


__license__ = 'GPL v3'
__copyright__ = '2015, Kovid Goyal <kovid at kovidgoyal.net>'

import numbers
from ctypes import POINTER, WINFUNCTYPE, byref, c_char_p, c_ulong, c_void_p, windll
from ctypes.wintypes import BOOL, DWORD, LPCWSTR, UINT

from polyglot.builtins import itervalues

HCONV     = c_void_p  # = DECLARE_HANDLE(HCONV)
HDDEDATA  = c_void_p  # = DECLARE_HANDLE(HDDEDATA)
HSZ       = c_void_p  # = DECLARE_HANDLE(HSZ)
LPBYTE    = c_char_p  # POINTER(BYTE)
LPDWORD   = POINTER(DWORD)
LPSTR    = c_char_p
ULONG_PTR = c_ulong

DML_ERRORS = {
    'ADVACKTIMEOUT': (0x4000, 'A request for a synchronous advise transaction has timed out.'),

    'BUSY': (0x4001, 'The response to the transaction caused the DDE_FBUSY flag to be set.'),

    'DATAACKTIMEOUT': (0x4002, 'A request for a synchronous data transaction has timed out.'),

    'DLL_NOT_INITIALIZED': (0x4003, 'A DDEML function was called without first calling the DdeInitialize function, or an invalid instance identifier was passed to a DDEML function.'),  # noqa: E501

    'DLL_USAGE': (0x4004, 'An application initialized as APPCLASS_MONITOR has attempted to perform a DDE transaction, or an application initialized as APPCMD_CLIENTONLY has attempted to perform server transactions.'),  # noqa: E501

    'EXECACKTIMEOUT': (0x4005, 'A request for a synchronous execute transaction has timed out.'),

    'INVALIDPARAMETER': (0x4006, 'An invalid transaction identifier was passed to a DDEML function. Once the application has returned from an XTYP_XACT_COMPLETE callback, the transaction identifier for that callback function is no longer valid.  A parameter failed to be validated by the DDEML. Some of the possible causes follow: The application used a data handle initialized with a different item name handle than was required by the transaction.  The application used a data handle that was initialized with a different clipboard data format than was required by the transaction.  The application used a client-side conversation handle with a server-side function or vice versa.  The application used a freed data handle or string handle.  More than one instance of the application used the same object.'),  # noqa: E501

    'LOW_MEMORY': (0x4007, 'A DDEML application has created a prolonged race condition (in which the server application outruns the client), causing large amounts of memory to be consumed.'),  # noqa: E501

    'MEMORY_ERROR': (0x4008, 'A memory allocation has failed.'),

    'NO_CONV_ESTABLISHED': (0x400a, "A client's attempt to establish a conversation has failed."),

    'NOTPROCESSED': (0x4009, 'A transaction has failed.'),

    'POKEACKTIMEOUT': (0x400b, 'A request for a synchronous poke transaction has timed out.'),

    'POSTMSG_FAILED': (0x400c, 'An internal call to the PostMessage function has failed.'),

    'REENTRANCY': (0x400d, 'An application instance with a synchronous transaction already in progress attempted to initiate another synchronous transaction, or the DdeEnableCallback function was called from within a DDEML callback function.'),  # noqa: E501

    'SERVER_DIED': (0x400e, 'A server-side transaction was attempted on a conversation terminated by the client, or the server terminated before completing a transaction.'),  # noqa: E501

    'SYS_ERROR': (0x400f, 'An internal error has occurred in the DDEML.'),

    'UNADVACKTIMEOUT': (0x4010, 'A request to end an advise transaction has timed out.'),

    'UNFOUND_QUEUE_ID': (0x4011, 'An invalid transaction identifier was passed to a DDEML function. Once the application has returned from an XTYP_XACT_COMPLETE callback, the transaction identifier for that callback function is no longer valid.'),  # noqa: E501
}
DML_ERROR_TEXT = dict(itervalues(DML_ERRORS))

user32 = windll.user32

DDECALLBACK = WINFUNCTYPE(HDDEDATA, UINT, UINT, HCONV, HSZ, HSZ, HDDEDATA, ULONG_PTR, ULONG_PTR)
APPCMD_CLIENTONLY = 0x10
CP_WINUNICODE = 1200
# See windows/ddeml.h for declaration of struct CONVCONTEXT
PCONVCONTEXT = c_void_p
XCLASS_FLAGS         = 0x4000
XTYP_EXECUTE         = (0x0050 | XCLASS_FLAGS)


class DDEError(ValueError):
    pass


def init_errcheck(result, func, args):
    if result != 0:
        raise DDEError(f'Failed to initialize DDE client with return code: {result:x}')
    return args


def no_errcheck(result, func, args):
    return args


def dde_error(instance):
    errcode = GetLastError(instance)
    raise DDEError(DML_ERRORS.get(errcode, f'Unknown DDE error code: {errcode:x}'))


def default_errcheck(result, func, args):
    if (isinstance(result, numbers.Integral) and result == 0) or (getattr(result, 'value', False) is None):
        dde_error(args[0])
    return args


null = object()


class a:

    def __init__(self, name, typ, default=null, in_arg=True):
        self.typ=typ
        if default is null:
            self.spec=((1 if in_arg else 2), name)
        else:
            self.spec=((1 if in_arg else 2), name, default)


def cwrap(name, restype, *args, **kw):
    params=(restype,) + tuple(x.typ for x in args)
    paramflags=tuple(x.spec for x in args)
    func=WINFUNCTYPE(*params)((name, kw.get('lib', user32)), paramflags)
    func.errcheck=kw.get('errcheck', default_errcheck)
    return func


GetLastError = cwrap('DdeGetLastError', UINT, a('instance', DWORD), errcheck=no_errcheck)

Initialize = cwrap('DdeInitializeW', UINT, a('instance_p', LPDWORD), a('callback', DDECALLBACK), a('command', DWORD),
                   a('reserved', DWORD, 0), errcheck=init_errcheck)
CreateStringHandle = cwrap('DdeCreateStringHandleW', HSZ, a('instance', DWORD), a('string', LPCWSTR), a('codepage', UINT, CP_WINUNICODE))
Connect = cwrap('DdeConnect', HCONV, a('instance', DWORD), a('service', HSZ), a('topic', HSZ), a('context', PCONVCONTEXT))
FreeStringHandle = cwrap('DdeFreeStringHandle', BOOL, a('instance', DWORD), a('handle', HSZ), errcheck=no_errcheck)
ClientTransaction = cwrap('DdeClientTransaction', HDDEDATA, a('data', LPBYTE), a('size', DWORD), a('conversation', HCONV), a('item', HSZ),
                          a('fmt', UINT, 0), a('type', UINT, XTYP_EXECUTE), a('timeout', DWORD, 5000), a('result', LPDWORD, LPDWORD()), errcheck=no_errcheck)
FreeDataHandle = cwrap('DdeFreeDataHandle', BOOL, a('data', HDDEDATA), errcheck=no_errcheck)
Disconnect = cwrap('DdeDisconnect', BOOL, a('conversation', HCONV), errcheck=no_errcheck)
Uninitialize = cwrap('DdeUninitialize', BOOL, a('instance', DWORD), errcheck=no_errcheck)


def send_dde_command(service, topic, command):
    instance = DWORD(0)

    def cb(*args):
        pass  # ignore callbacks
    callback = DDECALLBACK(cb)

    Initialize(byref(instance), callback, APPCMD_CLIENTONLY, 0)
    hservice = CreateStringHandle(instance, service)
    htopic = CreateStringHandle(instance, topic)
    conversation = Connect(instance, hservice, htopic, PCONVCONTEXT())
    FreeStringHandle(instance, hservice)
    FreeStringHandle(instance, htopic)
    data = c_char_p(command)
    sz = DWORD(len(command) + 1)
    res = ClientTransaction(data, sz, conversation, HSZ())
    if res == 0:
        dde_error(instance)
    FreeDataHandle(res)
    Disconnect(conversation)
    Uninitialize(instance)


if __name__ == '__main__':
    send_dde_command('WinWord', 'System', '[REM_DDE_Direct][FileOpen("C:/cygwin64/home/kovid/demo.docx")]')



#!/usr/local/bin/python3.13

__license__   = 'GPL v3'
__copyright__ = '2011, Kovid Goyal <kovid@kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

import numbers
import os
import re
import traceback
from collections import defaultdict
from contextlib import suppress
from copy import deepcopy
from functools import partial

from calibre.constants import CONFIG_DIR_MODE, config_dir, filesystem_encoding, get_umask, iswindows, preferred_encoding
from calibre.utils.localization import _
from calibre.utils.resources import get_path as P
from polyglot.builtins import iteritems

plugin_dir = os.path.join(config_dir, 'plugins')


def parse_old_style(src):
    import pickle as cPickle
    options = {'cPickle':cPickle}
    try:
        if not isinstance(src, str):
            src = src.decode('utf-8')
        src = re.sub(r'PyQt(?:4|5).QtCore', r'PyQt6.QtCore', src)
        src = re.sub(r'cPickle\.loads\(([\'"])', r'cPickle.loads(b\1', src)
        exec(src, options)
    except Exception as err:
        try:
            print(f'Failed to parse old style options string with error: {err}')
        except Exception:
            pass
    return options


def to_json(obj):
    import datetime
    if isinstance(obj, bytearray):
        from base64 import standard_b64encode
        return {'__class__': 'bytearray',
                '__value__': standard_b64encode(bytes(obj)).decode('ascii')}
    if isinstance(obj, datetime.datetime):
        from calibre.utils.date import isoformat
        return {'__class__': 'datetime.datetime',
                '__value__': isoformat(obj, as_utc=True)}
    if isinstance(obj, (set, frozenset)):
        return {'__class__': 'set', '__value__': tuple(obj)}
    if isinstance(obj, bytes):
        return obj.decode('utf-8')
    if hasattr(obj, 'toBase64'):  # QByteArray
        return {'__class__': 'bytearray',
                '__value__': bytes(obj.toBase64()).decode('ascii')}
    v = getattr(obj, 'value', None)
    if isinstance(v, int):  # Possibly an enum with integer values like all the Qt enums
        return v
    raise TypeError(repr(obj) + ' is not JSON serializable')


def safe_to_json(obj):
    try:
        return to_json(obj)
    except Exception:
        pass


def from_json(obj):
    custom = obj.get('__class__')
    if custom is not None:
        if custom == 'bytearray':
            from base64 import standard_b64decode
            return bytearray(standard_b64decode(obj['__value__'].encode('ascii')))
        if custom == 'datetime.datetime':
            from calibre.utils.iso8601 import parse_iso8601
            return parse_iso8601(obj['__value__'], assume_utc=True)
        if custom == 'set':
            return set(obj['__value__'])
    return obj


def force_unicode(x):
    try:
        return x.decode('mbcs' if iswindows else preferred_encoding)
    except UnicodeDecodeError:
        try:
            return x.decode(filesystem_encoding)
        except UnicodeDecodeError:
            return x.decode('utf-8', 'replace')


def force_unicode_recursive(obj):
    if isinstance(obj, bytes):
        return force_unicode(obj)
    if isinstance(obj, (list, tuple)):
        return type(obj)(map(force_unicode_recursive, obj))
    if isinstance(obj, dict):
        return {force_unicode_recursive(k): force_unicode_recursive(v) for k, v in iteritems(obj)}
    return obj


def json_dumps(obj, ignore_unserializable=False):
    import json
    try:
        ans = json.dumps(obj, indent=2, default=safe_to_json if ignore_unserializable else to_json, sort_keys=True, ensure_ascii=False)
    except UnicodeDecodeError:
        obj = force_unicode_recursive(obj)
        ans = json.dumps(obj, indent=2, default=safe_to_json if ignore_unserializable else to_json, sort_keys=True, ensure_ascii=False)
    if not isinstance(ans, bytes):
        ans = ans.encode('utf-8')
    return ans


def json_loads(raw):
    import json
    if isinstance(raw, bytes):
        raw = raw.decode('utf-8')
    return json.loads(raw, object_hook=from_json)


def make_config_dir():
    if not os.path.exists(plugin_dir):
        os.makedirs(plugin_dir, mode=CONFIG_DIR_MODE)


class Option:

    def __init__(self, name, switches=[], help='', type=None, choices=None,
                 check=None, group=None, default=None, action=None, metavar=None):
        if choices:
            type = 'choice'

        self.name     = name
        self.switches = switches
        self.help     = help.replace('%default', repr(default)) if help else None
        self.type     = type
        if self.type is None and action is None and choices is None:
            if isinstance(default, float):
                self.type = 'float'
            elif isinstance(default, numbers.Integral) and not isinstance(default, bool):
                self.type = 'int'

        self.choices  = choices
        self.check    = check
        self.group    = group
        self.default  = default
        self.action   = action
        self.metavar  = metavar

    def __eq__(self, other):
        return self.name == getattr(other, 'name', other)

    def __repr__(self):
        return 'Option: '+self.name

    def __str__(self):
        return repr(self)


class OptionValues:

    def copy(self):
        return deepcopy(self)


class OptionSet:

    OVERRIDE_PAT = re.compile(r'#{3,100} Override Options #{15}(.*?)#{3,100} End Override #{3,100}',
                              re.DOTALL|re.IGNORECASE)

    def __init__(self, description=''):
        self.description = description
        self.defaults = {}
        self.preferences = []
        self.group_list  = []
        self.groups      = {}
        self.set_buffer  = {}
        self.loads_pat = None

    def has_option(self, name_or_option_object):
        if name_or_option_object in self.preferences:
            return True
        for p in self.preferences:
            if p.name == name_or_option_object:
                return True
        return False

    def get_option(self, name_or_option_object):
        idx = self.preferences.index(name_or_option_object)
        if idx > -1:
            return self.preferences[idx]
        for p in self.preferences:
            if p.name == name_or_option_object:
                return p

    def add_group(self, name, description=''):
        if name in self.group_list:
            raise ValueError(f'A group by the name {name} already exists in this set')
        self.groups[name] = description
        self.group_list.append(name)
        return partial(self.add_opt, group=name)

    def update(self, other):
        for name in other.groups.keys():
            self.groups[name] = other.groups[name]
            if name not in self.group_list:
                self.group_list.append(name)
        for pref in other.preferences:
            if pref in self.preferences:
                self.preferences.remove(pref)
            self.preferences.append(pref)

    def smart_update(self, opts1, opts2):
        '''
        Updates the preference values in opts1 using only the non-default preference values in opts2.
        '''
        for pref in self.preferences:
            new = getattr(opts2, pref.name, pref.default)
            if new != pref.default:
                setattr(opts1, pref.name, new)

    def remove_opt(self, name):
        if name in self.preferences:
            self.preferences.remove(name)

    def add_opt(self, name, switches=[], help=None, type=None, choices=None,
                 group=None, default=None, action=None, metavar=None):
        '''
        Add an option to this section.

        :param name:       The name of this option. Must be a valid Python identifier.
                           Must also be unique in this OptionSet and all its subsets.
        :param switches:   List of command line switches for this option
                           (as supplied to :module:`optparse`). If empty, this
                           option will not be added to the command line parser.
        :param help:       Help text.
        :param type:       Type checking of option values. Supported types are:
                           `None, 'choice', 'complex', 'float', 'int', 'string'`.
        :param choices:    List of strings or `None`.
        :param group:      Group this option belongs to. You must previously
                           have created this group with a call to :method:`add_group`.
        :param default:    The default value for this option.
        :param action:     The action to pass to optparse. Supported values are:
                           `None, 'count'`. For choices and boolean options,
                           action is automatically set correctly.
        '''
        pref = Option(name, switches=switches, help=help, type=type, choices=choices,
                 group=group, default=default, action=action, metavar=None)
        if group is not None and group not in self.groups.keys():
            raise ValueError(f'Group {group} has not been added to this section')
        if pref in self.preferences:
            raise ValueError(f'An option with the name {name} already exists in this set.')
        self.preferences.append(pref)
        self.defaults[name] = default

    def retranslate_help(self):
        t = _
        for opt in self.preferences:
            if opt.help:
                opt.help = t(opt.help)
                if opt.name == 'use_primary_find_in_search':
                    opt.help = opt.help.format('ñ')

    def option_parser(self, user_defaults=None, usage='', gui_mode=False):
        from calibre.utils.config import OptionParser
        parser = OptionParser(usage, gui_mode=gui_mode)
        groups = defaultdict(lambda: parser)
        for group, desc in self.groups.items():
            groups[group] = parser.add_option_group(group.upper(), desc)

        for pref in self.preferences:
            if not pref.switches:
                continue
            g = groups[pref.group]
            action = pref.action
            if action is None:
                action = 'store'
                if pref.default is True or pref.default is False:
                    action = 'store_' + ('false' if pref.default else 'true')
            args = dict(
                        dest=pref.name,
                        help=pref.help,
                        metavar=pref.metavar,
                        type=pref.type,
                        choices=pref.choices,
                        default=getattr(user_defaults, pref.name, pref.default),
                        action=action,
                        )
            g.add_option(*pref.switches, **args)

        return parser

    def get_override_section(self, src):
        match = self.OVERRIDE_PAT.search(src)
        if match:
            return match.group()
        return ''

    def parse_string(self, src):
        options = {}
        if src:
            is_old_style = (isinstance(src, bytes) and src.startswith(b'#')) or (isinstance(src, str) and src.startswith('#'))
            if is_old_style:
                options = parse_old_style(src)
            else:
                try:
                    options = json_loads(src)
                    if not isinstance(options, dict):
                        raise Exception('options is not a dictionary')
                except Exception as err:
                    try:
                        print(f'Failed to parse options string with error: {err}')
                    except Exception:
                        pass
        opts = OptionValues()
        for pref in self.preferences:
            val = options.get(pref.name, pref.default)
            formatter = __builtins__.get(pref.type, None)
            if callable(formatter):
                val = formatter(val)
            setattr(opts, pref.name, val)

        return opts

    def serialize(self, opts, ignore_unserializable=False):
        data = {pref.name: getattr(opts, pref.name, pref.default) for pref in self.preferences}
        return json_dumps(data, ignore_unserializable=ignore_unserializable)


class ConfigInterface:

    def __init__(self, description):
        self.option_set       = OptionSet(description=description)
        self.add_opt          = self.option_set.add_opt
        self.add_group        = self.option_set.add_group
        self.remove_opt       = self.remove = self.option_set.remove_opt
        self.parse_string     = self.option_set.parse_string
        self.get_option       = self.option_set.get_option
        self.preferences      = self.option_set.preferences

    def update(self, other):
        self.option_set.update(other.option_set)

    def option_parser(self, usage='', gui_mode=False):
        return self.option_set.option_parser(user_defaults=self.parse(),
                                             usage=usage, gui_mode=gui_mode)

    def smart_update(self, opts1, opts2):
        self.option_set.smart_update(opts1, opts2)


def retry_on_fail(func, *args, count=10, sleep_time=0.2):
    import time
    ERROR_SHARING_VIOLATION = 32
    ACCESS_DENIED = 5
    for i in range(count):
        try:
            return func(*args)
        except FileNotFoundError:
            raise
        except OSError as e:
            # Windows stupidly gives us an ACCESS_DENIED rather than a
            # ERROR_SHARING_VIOLATION if the file is open
            if not iswindows or i > count - 2 or e.winerror not in (ERROR_SHARING_VIOLATION, ACCESS_DENIED):
                raise
            # Try the operation repeatedly in case something like a virus
            # scanner has opened one of the files (I love windows)
            time.sleep(sleep_time)


def read_data(file_path):
    def r():
        with open(file_path, 'rb') as f:
            return f.read()
    return retry_on_fail(r)


def commit_data(file_path, data):
    import tempfile
    bdir = os.path.dirname(file_path)
    os.makedirs(bdir, exist_ok=True, mode=CONFIG_DIR_MODE)
    try:
        with tempfile.NamedTemporaryFile(dir=bdir, prefix=os.path.basename(file_path).split('.')[0] + '-atomic-', delete=False) as f:
            if hasattr(os, 'fchmod'):
                os.fchmod(f.fileno(), 0o666 & ~get_umask())
            f.write(data)
        retry_on_fail(os.replace, f.name, file_path)
    finally:
        with suppress(FileNotFoundError, NameError):
            os.remove(f.name)


class Config(ConfigInterface):
    '''
    A file based configuration.
    '''

    def __init__(self, basename, description=''):
        ConfigInterface.__init__(self, description)
        self.filename_base = basename

    @property
    def config_file_path(self):
        return os.path.join(config_dir, self.filename_base + '.py.json')

    def parse(self):
        src = ''
        migrate = False
        path = self.config_file_path
        with suppress(FileNotFoundError):
            src_bytes = read_data(path)
            try:
                src = src_bytes.decode('utf-8')
            except ValueError:
                print('Failed to parse', path)
                traceback.print_exc()
        if not src:
            path = path.rpartition('.')[0]
            from calibre.utils.shared_file import share_open
            try:
                with share_open(path, 'rb') as f:
                    src = f.read().decode('utf-8')
            except Exception:
                pass
            else:
                migrate = bool(src)
        ans = self.option_set.parse_string(src)
        if migrate:
            new_src = self.option_set.serialize(ans, ignore_unserializable=True)
            commit_data(self.config_file_path, new_src)
        return ans

    def set(self, name, val):
        if not self.option_set.has_option(name):
            raise ValueError(f'The option {name} is not defined.')
        if not os.path.exists(config_dir):
            make_config_dir()
        src = b''
        with suppress(FileNotFoundError):
            src = read_data(self.config_file_path)
        opts = self.option_set.parse_string(src)
        setattr(opts, name, val)
        src = self.option_set.serialize(opts)
        if isinstance(src, str):
            src = src.encode('utf-8')
        commit_data(self.config_file_path, src)


class StringConfig(ConfigInterface):
    '''
    A string based configuration
    '''

    def __init__(self, src, description=''):
        ConfigInterface.__init__(self, description)
        self.set_src(src)

    def set_src(self, src):
        self.src = src
        if isinstance(self.src, bytes):
            self.src = self.src.decode('utf-8')

    def parse(self):
        return self.option_set.parse_string(self.src)

    def set(self, name, val):
        if not self.option_set.has_option(name):
            raise ValueError(f'The option {name} is not defined.')
        opts = self.option_set.parse_string(self.src)
        setattr(opts, name, val)
        self.set_src(self.option_set.serialize(opts))


class ConfigProxy:
    '''
    A Proxy to minimize file reads for widely used config settings
    '''

    def __init__(self, config):
        self.__config = config
        self.__opts   = None

    @property
    def defaults(self):
        return self.__config.option_set.defaults

    def refresh(self):
        self.__opts = self.__config.parse()

    def retranslate_help(self):
        self.__config.option_set.retranslate_help()

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, val):
        return self.set(key, val)

    def __delitem__(self, key):
        self.set(key, self.defaults[key])

    def get(self, key):
        if self.__opts is None:
            self.refresh()
        return getattr(self.__opts, key)

    def set(self, key, val):
        if self.__opts is None:
            self.refresh()
        setattr(self.__opts, key, val)
        return self.__config.set(key, val)

    def help(self, key):
        return self.__config.get_option(key).help


def create_global_prefs(conf_obj=None):
    c = Config('global', 'calibre wide preferences') if conf_obj is None else conf_obj
    c.add_opt('database_path',
              default=os.path.expanduser('~/library1.db'),
              help=_('Path to the database in which books are stored'))
    c.add_opt('filename_pattern', default='(?P<title>.+) - (?P<author>[^_]+)',
              help=_('Pattern to guess metadata from filenames'))
    c.add_opt('isbndb_com_key', default='',
              help=_('Access key for isbndb.com'))
    c.add_opt('network_timeout', default=5,
              help=_('Default timeout for network operations (seconds)'))
    c.add_opt('library_path', default=None,
              help=_('Path to folder in which your library of books is stored'))
    c.add_opt('language', default=None,
              help=_('The language in which to display the user interface'))
    c.add_opt('output_format', default='EPUB',
              help=_('The default output format for e-book conversions. When auto-converting'
                  ' to send to a device this can be overridden by individual device preferences.'
                  ' These can be changed by right clicking the device icon in calibre and'
                  ' choosing "Configure".'))
    c.add_opt('input_format_order', default=['EPUB', 'AZW3', 'MOBI', 'LIT', 'PRC',
        'FB2', 'HTML', 'HTM', 'XHTM', 'SHTML', 'XHTML', 'ZIP', 'DOCX', 'ODT', 'RTF', 'PDF',
        'TXT'],
              help=_('Ordered list of formats to prefer for input.'))
    c.add_opt('read_file_metadata', default=True,
              help=_('Read metadata from files'))
    c.add_opt('worker_process_priority', default='normal',
              help=_('The priority of worker processes. A higher priority '
                  'means they run faster and consume more resources. '
                  'Most tasks like conversion/news download/adding books/etc. '
                  'are affected by this setting.'))
    c.add_opt('swap_author_names', default=False,
            help=_('Swap author first and last names when reading metadata'))
    c.add_opt('add_formats_to_existing', default=False,
            help=_('Add new formats to existing book records'))
    c.add_opt('check_for_dupes_on_ctl', default=False,
            help=_('Check for duplicates when copying to another library'))
    c.add_opt('installation_uuid', default=None, help='Installation UUID')
    c.add_opt('new_book_tags', default=[], help=_('Tags to apply to books added to the library'))
    c.add_opt('mark_new_books', default=False, help=_(
        'Mark newly added books. The mark is a temporary mark that is automatically removed when calibre is restarted.'))

    # these are here instead of the gui preferences because calibredb and
    # calibre server can execute searches
    c.add_opt('saved_searches', default={}, help=_('List of named saved searches'))
    c.add_opt('user_categories', default={}, help=_('User-created Tag browser categories'))
    c.add_opt('manage_device_metadata', default='manual',
        help=_('How and when calibre updates metadata on the device.'))
    c.add_opt('limit_search_columns', default=False,
            help=_('When searching for text without using lookup '
            'prefixes, as for example, Red instead of title:Red, '
            'limit the columns searched to those named below.'))
    c.add_opt('limit_search_columns_to',
            default=['title', 'authors', 'tags', 'series', 'publisher'],
            help=_('Choose columns to be searched when not using prefixes, '
                'as for example, when searching for Red instead of '
                'title:Red. Enter a list of search/lookup names '
                'separated by commas. Only takes effect if you set the option '
                'to limit search columns above.'))
    c.add_opt('use_primary_find_in_search', default=True,
            help=_('Characters typed in the search box will match their '
                   'accented versions, based on the language you have chosen '
                   'for the calibre interface. For example, in '
                   'English, searching for n will match both {} and n, but if '
                   'your language is Spanish it will only match n. Note that '
                   'this is much slower than a simple search on very large '
                   'libraries. Also, this option will have no effect if you turn '
                   'on case-sensitive searching.'))
    c.add_opt('case_sensitive', default=False, help=_(
        'Make searches case-sensitive'))
    c.add_opt('numeric_collation', default=False,
            help=_('Recognize numbers inside text when sorting. Setting this '
                   'means that when sorting on text fields like title the text "Book 2"'
                   'will sort before the text "Book 100". Note that setting this '
                   'can cause problems with text that starts with numbers and is '
                   'a little slower.'))

    c.add_opt('migrated', default=False, help="For Internal use. Don't modify.")
    return c


prefs = ConfigProxy(create_global_prefs())
if prefs['installation_uuid'] is None:
    import uuid
    prefs['installation_uuid'] = str(uuid.uuid4())


# Read tweaks

def tweaks_file():
    return os.path.join(config_dir, 'tweaks.json')


def make_unicode(obj):
    if isinstance(obj, bytes):
        try:
            return obj.decode('utf-8')
        except UnicodeDecodeError:
            return obj.decode(preferred_encoding, errors='replace')
    if isinstance(obj, (list, tuple)):
        return list(map(make_unicode, obj))
    if isinstance(obj, dict):
        return {make_unicode(k): make_unicode(v) for k, v in iteritems(obj)}
    return obj


def normalize_tweak(val):
    if isinstance(val, (list, tuple)):
        return tuple(map(normalize_tweak, val))
    if isinstance(val, dict):
        return {k: normalize_tweak(v) for k, v in iteritems(val)}
    return val


def write_custom_tweaks(tweaks_dict):
    make_config_dir()
    tweaks_dict = make_unicode(tweaks_dict)
    changed_tweaks = {}
    default_tweaks = exec_tweaks(default_tweaks_raw())
    for key, cval in iteritems(tweaks_dict):
        if key in default_tweaks and normalize_tweak(cval) == normalize_tweak(default_tweaks[key]):
            continue
        changed_tweaks[key] = cval
    raw = json_dumps(changed_tweaks)
    with open(tweaks_file(), 'wb') as f:
        f.write(raw)


def exec_tweaks(path):
    if isinstance(path, bytes):
        raw = path
        fname = '<string>'
    else:
        with open(path, 'rb') as f:
            raw = f.read()
            fname = f.name
    code = compile(raw, fname, 'exec')
    l = {}
    g = {'__file__': fname}
    exec(code, g, l)
    return l


def read_custom_tweaks():
    make_config_dir()
    tf = tweaks_file()
    ans = {}
    if os.path.exists(tf):
        with open(tf, 'rb') as f:
            raw = f.read()
        raw = raw.strip()
        if not raw:
            return ans
        try:
            return json_loads(raw)
        except Exception:
            import traceback
            traceback.print_exc()
            return ans
    old_tweaks_file = tf.rpartition('.')[0] + '.py'
    if os.path.exists(old_tweaks_file):
        ans = exec_tweaks(old_tweaks_file)
        ans = make_unicode(ans)
        write_custom_tweaks(ans)
    return ans


def default_tweaks_raw():
    return P('default_tweaks.py', data=True, allow_user_override=False)


def read_tweaks():
    default_tweaks = exec_tweaks(default_tweaks_raw())
    try:
        custom_tweaks = read_custom_tweaks()
    except Exception:
        custom_tweaks = {}
    default_tweaks.update(custom_tweaks)
    return default_tweaks


tweaks = read_tweaks()


def migrate_tweaks_to_prefs():
    # This must happen after the tweaks are loaded
    # Migrate the numeric_collation tweak
    if 'numeric_collation' in tweaks:
        prefs['numeric_collation'] = tweaks.get('numeric_collation', False)
        tweaks.pop('numeric_collation')
        write_custom_tweaks(tweaks)


migrate_tweaks_to_prefs()


def reset_tweaks_to_default():
    default_tweaks = exec_tweaks(default_tweaks_raw())
    tweaks.clear()
    tweaks.update(default_tweaks)


class Tweak:

    def __init__(self, name, value):
        self.name, self.value = name, value

    def __enter__(self):
        self.origval = tweaks[self.name]
        tweaks[self.name] = self.value

    def __exit__(self, *args):
        tweaks[self.name] = self.origval



#!/usr/local/bin/python3.13
# License: GPL v3 Copyright: 2019, Kovid Goyal <kovid at kovidgoyal.net>

import io
import sys
import time

from polyglot.builtins import as_bytes, as_unicode


def is_binary(stream):
    mode = getattr(stream, 'mode', None)
    if mode:
        return 'b' in mode
    return not isinstance(stream, io.TextIOBase)


def prints(*a, **kw):
    ' Print either unicode or bytes to either binary or text mode streams '
    stream = kw.get('file', sys.stdout)
    if stream is None:
        return
    sep, end = kw.get('sep'), kw.get('end')
    if sep is None:
        sep = ' '
    if end is None:
        end = '\n'
    if is_binary(stream):
        encoding = getattr(stream, 'encoding', None) or 'utf-8'
        a = (as_bytes(x, encoding=encoding) for x in a)
        sep = as_bytes(sep)
        end = as_bytes(end)
    else:
        a = (as_unicode(x, errors='replace') for x in a)
        sep = as_unicode(sep)
        end = as_unicode(end)
    for i, x in enumerate(a):
        if sep and i != 0:
            stream.write(sep)
        stream.write(x)
    if end:
        stream.write(end)
    if kw.get('flush'):
        try:
            stream.flush()
        except Exception:
            pass


def debug_print(*args, **kw):
    '''
    Prints debug information to the console if debugging is enabled.

    This function prints a message prefixed with a timestamp showing the elapsed time
    since the first call to this function. The message is printed only if debugging is enabled.

    Parameters:
    *args : tuple
        Variable length argument list to be printed.
    **kw : dict
        Arbitrary keyword arguments to be passed to the `print` function.

    Attributes:
    base_time : float
        The timestamp of the first call to this function. Stored as an attribute of the function.

    Behavior:
    - On the first call, initializes `base_time` to the current time using `time.monotonic()`.
    - If `is_debugging()` returns True, prints the elapsed time since `base_time` along with the provided arguments.
    '''
    from calibre.constants import is_debugging

    # Get the base_time attribute, initializing it on the first call
    base_time = getattr(debug_print, 'base_time', None)
    if base_time is None:
        # Set base_time to the current monotonic time if it hasn't been set
        debug_print.base_time = base_time = time.monotonic()

    # Check if debugging is enabled
    if is_debugging():
        # Print the elapsed time and the provided arguments if debugging is enabled
        prints(f'DEBUG: {time.monotonic()-base_time:6.1f}', *args, **kw)



#!/usr/local/bin/python3.13
# License: GPLv3 Copyright: 2022, Charles Haley

from enum import Enum
from functools import partial

from qt.core import QComboBox, QDialog, QDialogButtonBox, QFormLayout, QIcon, QLabel, QMenu, QToolButton, QVBoxLayout

from calibre.gui2 import error_dialog, gprefs, question_dialog
from calibre.gui2.actions import InterfaceAction, show_menu_under_widget
from calibre.gui2.geometry import _restore_geometry, delete_geometry, save_geometry
from calibre.utils.icu import sort_key


class Panel(Enum):
    ' See gui2.init for these '
    SEARCH_BAR = 'sb'
    TAG_BROWSER = 'tb'
    BOOK_DETAILS = 'bd'
    GRID_VIEW = 'gv'
    COVER_BROWSER = 'cb'
    QUICKVIEW = 'qv'


class SaveLayoutDialog(QDialog):

    def __init__(self, parent, names):
        QDialog.__init__(self, parent)
        self.names = names
        l = QVBoxLayout(self)
        fl = QFormLayout()
        l.addLayout(fl)
        self.cb = cb = QComboBox()
        cb.setEditable(True)
        cb.setMinimumWidth(200)
        cb.addItem('')
        cb.addItems(sorted(names, key=sort_key))
        fl.addRow(QLabel(_('Layout name')), cb)
        bb = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        l.addWidget(bb)
        bb.accepted.connect(self.accept)
        bb.rejected.connect(self.reject)

    def current_name(self):
        return self.cb.currentText().strip()

    def accept(self):
        n = self.current_name()
        if not n:
            error_dialog(self, _('Invalid name'), _('The settings name cannot be blank'),
                         show=True, show_copy_button=False)
            return
        if self.current_name() in self.names:
            r = question_dialog(self, _('Replace saved layout'),
                  _('Do you really want to overwrite the saved layout {0}?').format(self.current_name()))
            if r == QDialog.DialogCode.Accepted:
                super().accept()
            else:
                return
        super().accept()


class LayoutActions(InterfaceAction):

    name = 'Layout Actions'
    action_spec = (_('Layout actions'), 'layout.png',
                   _("Save and restore layout item sizes, and add/remove/toggle "
                     "layout items such as the search bar, tag browser, etc. "
                     "Item sizes in saved layouts are saved as a percentage of "
                     "the window size. Restoring a layout doesn't change the "
                     "window size, instead fitting the items into the current window."), None)

    action_type = 'current'
    popup_type = QToolButton.ToolButtonPopupMode.InstantPopup
    action_add_menu = True
    dont_add_to = frozenset({'context-menu-device', 'menubar-device'})

    def genesis(self):
        self.layout_icon = QIcon.ic('layout.png')
        self.menu = m = self.qaction.menu()
        m.aboutToShow.connect(self.about_to_show_menu)

        # Create a "hidden" menu that can have a shortcut.
        self.hidden_menu = QMenu()
        self.shortcut_action = self.create_menu_action(
                        menu=self.hidden_menu,
                        unique_name='Main window layout',
                        shortcut=None,
                        text=_('Save and restore layout item sizes, and add/remove/toggle '
                               'layout items such as the search bar, tag browser, etc. '),
                        icon='layout.png',
                        triggered=self.show_menu)

    # We want to show the menu when a shortcut is used. Apparently the only way
    # to do that is to scan the toolbar(s) for the action button then exec the
    # associated menu. The search is done here to take adding and removing the
    # action from toolbars into account.
    #
    # If a shortcut is triggered and there isn't a toolbar button visible then
    # show the menu in the upper left corner of the library view pane. Yes, this
    # is a bit weird but it works as well as a popping up a dialog.
    def show_menu(self):
        show_menu_under_widget(self.gui, self.menu, self.qaction, self.name)

    def toggle_layout(self):
        self.gui.layout_container.toggle_layout()

    def gui_layout_complete(self):
        m = self.qaction.menu()
        m.aboutToShow.connect(self.about_to_show_menu)

    def initialization_complete(self):
        self.populate_menu()

    def about_to_show_menu(self):
        self.populate_menu()

    def populate_menu(self):
        m = self.qaction.menu()
        m.clear()
        lm = m.addMenu(self.layout_icon, _('Restore saved layout'))
        layouts = gprefs['saved_layouts']
        if layouts:
            for l in sorted(layouts, key=sort_key):
                lm.addAction(self.layout_icon, l, partial(self.apply_layout, l))
        else:
            lm.setEnabled(False)
        lm = m.addAction(self.layout_icon, _('Save current layout'))
        lm.triggered.connect(self.save_current_layout)
        lm = m.addMenu(self.layout_icon, _('Delete saved layout'))
        layouts = gprefs['saved_layouts']
        if layouts:
            for l in sorted(layouts, key=sort_key):
                lm.addAction(self.layout_icon, l, partial(self.delete_layout, l))
        else:
            lm.setEnabled(False)

        m.addSeparator()
        m.addAction(_('Hide all'), self.hide_all)
        for button, name in zip(self.gui.layout_buttons, self.gui.button_order):
            m.addSeparator()
            ic = button.icon()
            m.addAction(ic, _('Show {}').format(button.label), partial(self.set_visible, Panel(name), True))
            m.addAction(ic, _('Hide {}').format(button.label), partial(self.set_visible, Panel(name), False))
            m.addAction(ic, _('Toggle {}').format(button.label), partial(self.toggle_item, Panel(name)))

    def _change_item(self, button, show=True):
        if button.isChecked() and not show:
            button.click()
        elif not button.isChecked() and show:
            button.click()

    def _toggle_item(self, button):
        button.click()

    def _button_from_enum(self, name: Panel):
        for q, b in zip(self.gui.button_order, self.gui.layout_buttons):
            if q == name.value:
                return b

    # Public API
    def apply_layout(self, name):
        '''apply_layout()
        Apply a saved GUI panel layout.

        :param:`name` The name of the saved layout

         Throws KeyError if the name doesn't exist.
        '''
        # This can be called by plugins so let the exception fly

        # Restore the application window geometry if we have it.
        _restore_geometry(self.gui, gprefs, f'saved_layout_{name}')
        # Now the panel sizes inside the central widget
        layouts = gprefs['saved_layouts']
        settings = layouts[name]
        # Order is important here. change_layout() must be called before
        # unserializing the settings or panes like book details won't display
        # properly.
        self.gui.layout_container.change_layout(self.gui, settings['layout'] == 'wide')
        self.gui.layout_container.unserialize_settings(settings)
        self.gui.layout_container.relayout()

    def save_current_layout(self):
        '''save_current_layout()
        Opens a dialog asking for the name to use to save the current layout.
        Saves the current settings under the provided name.
        '''
        layouts = gprefs['saved_layouts']
        d = SaveLayoutDialog(self.gui, layouts.keys())
        if d.exec() == QDialog.DialogCode.Accepted:
            self.save_named_layout(d.current_name(), self.current_settings())

    def current_settings(self):
        '''current_settings()

        :return: the current gui layout settings.
        '''

        return self.gui.layout_container.serialized_settings()

    def save_named_layout(self, name, settings):
        '''save_named_layout()
        Saves the settings under the provided name.

        :param:`name` The name for the settings.
        :param:`settings`: The gui layout settings to save.
        '''
        # Save the main window geometry.
        save_geometry(self.gui, gprefs, f'saved_layout_{name}')
        # Now the panel sizes inside the central widget
        layouts = gprefs['saved_layouts']
        layouts.update({name: settings})
        gprefs['saved_layouts'] = layouts
        self.populate_menu()

    def delete_layout(self, name, show_warning=True):
        '''delete_layout()
        Delete a saved layout.

        :param:`name` The name of the layout to delete
        :param:`show_warning`: If True a warning dialog will be shown before deleting the layout.
        '''
        if show_warning:
            if not question_dialog(self.gui, _('Are you sure?'),
                                   _('Do you really want to delete the saved layout {0}?').format(name),
                                   skip_dialog_name='delete_saved_gui_layout'):
                return

        # The information is stored as 2 preferences. Delete them both.
        delete_geometry(gprefs, f'saved_layout_{name}')
        layouts = gprefs['saved_layouts']
        layouts.pop(name, None)
        self.populate_menu()

    def saved_layout_names(self):
        '''saved_layout_names()
        Get a list of saved layout names

        :return: the sorted list of names. The list is empty if there are no names.
        '''
        layouts = gprefs['saved_layouts']
        return sorted(layouts.keys(), key=sort_key)

    def toggle_item(self, name):
        '''toggle_item()
        Toggle the visibility of the panel.

        :param name: specifies which panel to toggle. Valid names are
            SEARCH_BAR: 'sb'
            TAG_BROWSER: 'tb'
            BOOK_DETAILS: 'bd'
            GRID_VIEW: 'gv'
            COVER_BROWSER: 'cb'
            QUICKVIEW: 'qv'
        '''
        self._toggle_item(self._button_from_enum(name))

    def set_visible(self, name: Panel, show=True):
        '''set_visible()
        Show or hide a panel. Does nothing if the panel is already in the
        desired state.

        :param name: specifies which panel to show.  Valid names are
            SEARCH_BAR: 'sb'
            TAG_BROWSER: 'tb'
            BOOK_DETAILS: 'bd'
            GRID_VIEW: 'gv'
            COVER_BROWSER: 'cb'
            QUICKVIEW: 'qv'
        :param show: If True, show the panel, otherwise hide the panel
        '''
        self._change_item(self._button_from_enum(name), show)

    def is_visible(self, name: Panel):
        '''is_visible()
        Returns True if the panel is visible.

        :param name: specifies which panel. Valid names are
            SEARCH_BAR: 'sb'
            TAG_BROWSER: 'tb'
            BOOK_DETAILS: 'bd'
            GRID_VIEW: 'gv'
            COVER_BROWSER: 'cb'
            QUICKVIEW: 'qv'
        '''
        self._button_from_enum(name).isChecked()

    def hide_all(self):
        for name in self.gui.button_order:
            self.set_visible(Panel(name), show=False)

    def show_all(self):
        for name in self.gui.button_order:
            self.set_visible(Panel(name), show=True)

    def panel_titles(self):
        '''panel_titles()
        Return a dictionary of Panel Enum items to translated human readable title.
        Simplifies building dialogs, for example combo boxes of all the panel
        names or check boxes for each panel.

        :return: {Panel_enum_value: human readable title, ...}
        '''
        return {p: self._button_from_enum(p).label for p in Panel}



#!/usr/local/bin/python3.13
# License: GPL v3 Copyright: 2022, Kovid Goyal <kovid at kovidgoyal.net>


import os
import subprocess
import sys
import traceback
from contextlib import suppress
from queue import Queue
from threading import Event, Thread
from time import monotonic

from calibre import detect_ncpus, human_readable
from calibre.utils.ipc.simple_worker import start_pipe_worker

check_for_work = object()
quit = object()


class Job:

    def __init__(self, book_id, fmt, path, fmt_size, fmt_hash, start_time):
        self.book_id = book_id
        self.fmt = fmt
        self.fmt_size = fmt_size
        self.fmt_hash = fmt_hash
        self.path = path
        self.start_time = start_time


class Result:

    def __init__(self, job, err_msg=''):
        self.book_id = job.book_id
        self.fmt = job.fmt
        self.fmt_size = job.fmt_size
        self.fmt_hash = job.fmt_hash
        self.ok = not bool(err_msg)
        self.start_time = job.start_time
        if self.ok:
            with open(job.path + '.txt', 'rb') as src:
                try:
                    self.text = src.read().decode('utf-8', 'replace')
                except Exception:
                    self.ok = False
                    self.text = traceback.format_exc()
        else:
            self.text = err_msg


class Worker(Thread):

    code_to_exec = 'from calibre.db.fts.text import main; main({!r})'
    max_duration = 30  # minutes
    poll_interval = 0.1  # seconds

    def __init__(self, jobs_queue, supervise_queue):
        super().__init__(name='FTSWorker', daemon=True)
        self.currently_working = False
        self.jobs_queue = jobs_queue
        self.supervise_queue = supervise_queue
        self.keep_going = True
        self.working = False

    def run(self):
        while self.keep_going:
            x = self.jobs_queue.get()
            if x is quit:
                break
            self.working = True
            try:
                res = self.run_job(x)
                if res is not None and self.keep_going:
                    self.supervise_queue.put(res)
            except Exception:
                tb = traceback.format_exc()
                traceback.print_exc()
                if self.keep_going:
                    self.supervise_queue.put(Result(x, tb))
            finally:
                self.working = False

    def run_job(self, job):
        time_limit = monotonic() + (self.max_duration * 60)
        txtpath = job.path + '.txt'
        errpath = job.path + '.error'
        try:
            with open(errpath, 'wb') as error:
                p = start_pipe_worker(
                    self.code_to_exec.format(job.path),
                    stdout=subprocess.DEVNULL, stderr=error, stdin=subprocess.DEVNULL, priority='low',
                )
                while self.keep_going and monotonic() <= time_limit:
                    with suppress(subprocess.TimeoutExpired):
                        p.wait(self.poll_interval)
                        break
                if p.returncode is None:
                    p.kill()
                    if not self.keep_going:
                        return
                    return Result(job, _('Extracting text from the {0} file of size {1} took too long').format(
                        job.fmt, human_readable(job.fmt_size)))
                if os.path.exists(txtpath):
                    return Result(job)
            with open(errpath, 'rb') as f:
                err = f.read().decode('utf-8', 'replace')
                return Result(job, err)
        finally:
            with suppress(OSError):
                os.remove(job.path)
            with suppress(OSError):
                os.remove(txtpath)
            with suppress(OSError):
                os.remove(errpath)


class Pool:

    def __init__(self, dbref):
        self.max_workers = 1
        self.jobs_queue = Queue()
        self.supervise_queue = Queue()
        self.workers = []
        self.initialized = Event()
        self.dbref = dbref
        self.keep_going = True

    def initialize(self):
        if not self.initialized.is_set():
            self.supervisor_thread = Thread(name='FTSSupervisor', daemon=True, target=self.supervise)
            self.supervisor_thread.start()
            self.expand_workers()
            self.initialized.set()

    def prune_dead_workers(self):
        self.workers = [w for w in self.workers if w.is_alive()]

    def expand_workers(self):
        self.prune_dead_workers()
        while len(self.workers) < self.max_workers:
            self.workers.append(self.create_worker())

    def create_worker(self):
        w = Worker(self.jobs_queue, self.supervise_queue)
        w.start()
        return w

    def shrink_workers(self):
        self.prune_dead_workers()
        extra = len(self.workers) - self.max_workers
        while extra > 0:
            self.jobs_queue.put(quit)
            extra -= 1

    # external API {{{
    @property
    def num_of_workers(self):
        return len(self.workers)

    @num_of_workers.setter
    def num_of_workers(self, num):
        self.initialize()
        self.prune_dead_workers()
        num = min(max(1, num), detect_ncpus())
        if num != self.max_workers:
            self.max_workers = num
            if num > len(self.workers):
                self.expand_workers()
            elif num < len(self.workers):
                self.shrink_workers()

    @property
    def num_of_idle_workers(self):
        return sum(0 if w.working else 1 for w in self.workers)

    def check_for_work(self):
        self.initialize()
        self.supervise_queue.put(check_for_work)

    def add_job(self, book_id, fmt, path, fmt_size, fmt_hash, start_time):
        self.initialize()
        job = Job(book_id, fmt, path, fmt_size, fmt_hash, start_time)
        self.jobs_queue.put(job)

    def commit_result(self, result):
        text = result.text
        err_msg = ''
        if not result.ok:
            print(f'Failed to get text from book_id: {result.book_id} format: {result.fmt}', file=sys.stderr)
            print(text, file=sys.stderr)
            err_msg = text
            text = ''
        db = self.dbref()
        if db is not None:
            db.commit_fts_result(result.book_id, result.fmt, result.fmt_size, result.fmt_hash, text, err_msg, result.start_time)

    def shutdown(self):
        if self.initialized.is_set():
            self.keep_going = False
            for i in range(2):
                self.supervise_queue.put(quit)
            for w in self.workers:
                w.keep_going = False
                for i in range(2*len(self.workers)):
                    self.jobs_queue.put(quit)
            self.initialized.clear()

    def join(self):
        with suppress(AttributeError):
            self.supervisor_thread.join()
        for w in self.workers:
            w.join()
        self.workers = []
    # }}}

    def do_check_for_work(self):
        db = self.dbref()
        if db is not None:
            db.queue_next_fts_job()

    def supervise(self):
        while self.keep_going:
            x = self.supervise_queue.get()
            try:
                if x is check_for_work:
                    self.do_check_for_work()
                elif x is quit:
                    break
                elif isinstance(x, Result):
                    self.commit_result(x)
                    self.do_check_for_work()
            except Exception:
                traceback.print_exc()



__license__   = 'GPL v3'
__copyright__ = '2008, Marshall T. Vandegrift <llasram@gmail.com>'

'''
Microsoft LIT HTML tag and attribute tables, copied from ConvertLIT.
'''

TAGS = [
    None,
    None,
    None,
    'a',
    'acronym',
    'address',
    'applet',
    'area',
    'b',
    'base',
    'basefont',
    'bdo',
    'bgsound',
    'big',
    'blink',
    'blockquote',
    'body',
    'br',
    'button',
    'caption',
    'center',
    'cite',
    'code',
    'col',
    'colgroup',
    None,
    None,
    'dd',
    'del',
    'dfn',
    'dir',
    'div',
    'dl',
    'dt',
    'em',
    'embed',
    'fieldset',
    'font',
    'form',
    'frame',
    'frameset',
    None,
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'head',
    'hr',
    'html',
    'i',
    'iframe',
    'img',
    'input',
    'ins',
    'kbd',
    'label',
    'legend',
    'li',
    'link',
    'tag61',
    'map',
    'tag63',
    'tag64',
    'meta',
    'nextid',
    'nobr',
    'noembed',
    'noframes',
    'noscript',
    'object',
    'ol',
    'option',
    'p',
    'param',
    'plaintext',
    'pre',
    'q',
    'rp',
    'rt',
    'ruby',
    's',
    'samp',
    'script',
    'select',
    'small',
    'span',
    'strike',
    'strong',
    'style',
    'sub',
    'sup',
    'table',
    'tbody',
    'tc',
    'td',
    'textarea',
    'tfoot',
    'th',
    'thead',
    'title',
    'tr',
    'tt',
    'u',
    'ul',
    'var',
    'wbr',
    None,
    ]

ATTRS0 = {
    0x8010: 'tabindex',
    0x8046: 'title',
    0x804b: 'style',
    0x804d: 'disabled',
    0x83ea: 'class',
    0x83eb: 'id',
    0x83fe: 'datafld',
    0x83ff: 'datasrc',
    0x8400: 'dataformatas',
    0x87d6: 'accesskey',
    0x9392: 'lang',
    0x93ed: 'language',
    0x93fe: 'dir',
    0x9771: 'onmouseover',
    0x9772: 'onmouseout',
    0x9773: 'onmousedown',
    0x9774: 'onmouseup',
    0x9775: 'onmousemove',
    0x9776: 'onkeydown',
    0x9777: 'onkeyup',
    0x9778: 'onkeypress',
    0x9779: 'onclick',
    0x977a: 'ondblclick',
    0x977e: 'onhelp',
    0x977f: 'onfocus',
    0x9780: 'onblur',
    0x9783: 'onrowexit',
    0x9784: 'onrowenter',
    0x9786: 'onbeforeupdate',
    0x9787: 'onafterupdate',
    0x978a: 'onreadystatechange',
    0x9790: 'onscroll',
    0x9794: 'ondragstart',
    0x9795: 'onresize',
    0x9796: 'onselectstart',
    0x9797: 'onerrorupdate',
    0x9799: 'ondatasetchanged',
    0x979a: 'ondataavailable',
    0x979b: 'ondatasetcomplete',
    0x979c: 'onfilterchange',
    0x979f: 'onlosecapture',
    0x97a0: 'onpropertychange',
    0x97a2: 'ondrag',
    0x97a3: 'ondragend',
    0x97a4: 'ondragenter',
    0x97a5: 'ondragover',
    0x97a6: 'ondragleave',
    0x97a7: 'ondrop',
    0x97a8: 'oncut',
    0x97a9: 'oncopy',
    0x97aa: 'onpaste',
    0x97ab: 'onbeforecut',
    0x97ac: 'onbeforecopy',
    0x97ad: 'onbeforepaste',
    0x97af: 'onrowsdelete',
    0x97b0: 'onrowsinserted',
    0x97b1: 'oncellchange',
    0x97b2: 'oncontextmenu',
    0x97b6: 'onbeforeeditfocus',
    }
ATTRS3 = {
    0x0001: 'href',
    0x03ec: 'target',
    0x03ee: 'rel',
    0x03ef: 'rev',
    0x03f0: 'urn',
    0x03f1: 'methods',
    0x8001: 'name',
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS5 = {
    0x9399: 'clear',
    }
ATTRS6 = {
    0x8001: 'name',
    0x8006: 'width',
    0x8007: 'height',
    0x804a: 'align',
    0x8bbb: 'classid',
    0x8bbc: 'data',
    0x8bbf: 'codebase',
    0x8bc0: 'codetype',
    0x8bc1: 'code',
    0x8bc2: 'type',
    0x8bc5: 'vspace',
    0x8bc6: 'hspace',
    0x978e: 'onerror',
    }
ATTRS7 = {
    0x0001: 'href',
    0x03ea: 'shape',
    0x03eb: 'coords',
    0x03ed: 'target',
    0x03ee: 'alt',
    0x03ef: 'nohref',
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS8 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS9 = {
    0x03ec: 'href',
    0x03ed: 'target',
    }
ATTRS10 = {
    0x938b: 'color',
    0x939b: 'face',
    0x93a3: 'size',
    }
ATTRS12 = {
    0x03ea: 'src',
    0x03eb: 'loop',
    0x03ec: 'volume',
    0x03ed: 'balance',
    }
ATTRS13 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS15 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x9399: 'clear',
    }
ATTRS16 = {
    0x07db: 'link',
    0x07dc: 'alink',
    0x07dd: 'vlink',
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x938a: 'background',
    0x938b: 'text',
    0x938e: 'nowrap',
    0x93ae: 'topmargin',
    0x93af: 'rightmargin',
    0x93b0: 'bottommargin',
    0x93b1: 'leftmargin',
    0x93b6: 'bgproperties',
    0x93d8: 'scroll',
    0x977b: 'onselect',
    0x9791: 'onload',
    0x9792: 'onunload',
    0x9798: 'onbeforeunload',
    0x97b3: 'onbeforeprint',
    0x97b4: 'onafterprint',
    0xfe0c: 'bgcolor',
    }
ATTRS17 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x9399: 'clear',
    }
ATTRS18 = {
    0x07d1: 'type',
    0x8001: 'name',
    }
ATTRS19 = {
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x93a8: 'valign',
    }
ATTRS20 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x9399: 'clear',
    }
ATTRS21 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS22 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS23 = {
    0x03ea: 'span',
    0x8006: 'width',
    0x8049: 'align',
    0x93a8: 'valign',
    0xfe0c: 'bgcolor',
    }
ATTRS24 = {
    0x03ea: 'span',
    0x8006: 'width',
    0x8049: 'align',
    0x93a8: 'valign',
    0xfe0c: 'bgcolor',
    }
ATTRS27 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x938e: 'nowrap',
    }
ATTRS29 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS31 = {
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x938e: 'nowrap',
    }
ATTRS32 = {
    0x03ea: 'compact',
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS33 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x938e: 'nowrap',
    }
ATTRS34 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS35 = {
    0x8001: 'name',
    0x8006: 'width',
    0x8007: 'height',
    0x804a: 'align',
    0x8bbd: 'palette',
    0x8bbe: 'pluginspage',
    # 0x8bbf: "codebase",
    0x8bbf: 'src',
    0x8bc1: 'units',
    0x8bc2: 'type',
    0x8bc3: 'hidden',
    }
ATTRS36 = {
    0x804a: 'align',
    }
ATTRS37 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x938b: 'color',
    0x939b: 'face',
    0x939c: 'size',
    }
ATTRS38 = {
    0x03ea: 'action',
    0x03ec: 'enctype',
    0x03ed: 'method',
    0x03ef: 'target',
    0x03f4: 'accept-charset',
    0x8001: 'name',
    0x977c: 'onsubmit',
    0x977d: 'onreset',
    }
ATTRS39 = {
    0x8000: 'align',
    0x8001: 'name',
    0x8bb9: 'src',
    0x8bbb: 'border',
    0x8bbc: 'frameborder',
    0x8bbd: 'framespacing',
    0x8bbe: 'marginwidth',
    0x8bbf: 'marginheight',
    0x8bc0: 'noresize',
    0x8bc1: 'scrolling',
    0x8fa2: 'bordercolor',
    }
ATTRS40 = {
    0x03e9: 'rows',
    0x03ea: 'cols',
    0x03eb: 'border',
    0x03ec: 'bordercolor',
    0x03ed: 'frameborder',
    0x03ee: 'framespacing',
    0x8001: 'name',
    0x9791: 'onload',
    0x9792: 'onunload',
    0x9798: 'onbeforeunload',
    0x97b3: 'onbeforeprint',
    0x97b4: 'onafterprint',
    }
ATTRS42 = {
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x9399: 'clear',
    }
ATTRS43 = {
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x9399: 'clear',
    }
ATTRS44 = {
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x9399: 'clear',
    }
ATTRS45 = {
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x9399: 'clear',
    }
ATTRS46 = {
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x9399: 'clear',
    }
ATTRS47 = {
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x9399: 'clear',
    }
ATTRS49 = {
    0x03ea: 'noshade',
    0x8006: 'width',
    0x8007: 'size',
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x938b: 'color',
    }
ATTRS51 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS52 = {
    0x8001: 'name',
    0x8006: 'width',
    0x8007: 'height',
    0x804a: 'align',
    0x8bb9: 'src',
    0x8bbb: 'border',
    0x8bbc: 'frameborder',
    0x8bbd: 'framespacing',
    0x8bbe: 'marginwidth',
    0x8bbf: 'marginheight',
    0x8bc0: 'noresize',
    0x8bc1: 'scrolling',
    0x8fa2: 'vspace',
    0x8fa3: 'hspace',
    }
ATTRS53 = {
    0x03eb: 'alt',
    0x03ec: 'src',
    0x03ed: 'border',
    0x03ee: 'vspace',
    0x03ef: 'hspace',
    0x03f0: 'lowsrc',
    0x03f1: 'vrml',
    0x03f2: 'dynsrc',
    0x03f4: 'loop',
    0x03f6: 'start',
    0x07d3: 'ismap',
    0x07d9: 'usemap',
    0x8001: 'name',
    0x8006: 'width',
    0x8007: 'height',
    0x8046: 'title',
    0x804a: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x978d: 'onabort',
    0x978e: 'onerror',
    0x9791: 'onload',
    }
ATTRS54 = {
    0x07d1: 'type',
    0x07d3: 'size',
    0x07d4: 'maxlength',
    0x07d6: 'readonly',
    0x07d8: 'indeterminate',
    0x07da: 'checked',
    0x07db: 'alt',
    0x07dc: 'src',
    0x07dd: 'border',
    0x07de: 'vspace',
    0x07df: 'hspace',
    0x07e0: 'lowsrc',
    0x07e1: 'vrml',
    0x07e2: 'dynsrc',
    0x07e4: 'loop',
    0x07e5: 'start',
    0x8001: 'name',
    0x8006: 'width',
    0x8007: 'height',
    0x804a: 'align',
    0x93ee: 'value',
    0x977b: 'onselect',
    0x978d: 'onabort',
    0x978e: 'onerror',
    0x978f: 'onchange',
    0x9791: 'onload',
    }
ATTRS56 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS57 = {
    0x03e9: 'for',
    }
ATTRS58 = {
    0x804a: 'align',
    }
ATTRS59 = {
    0x03ea: 'value',
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x939a: 'type',
    }
ATTRS60 = {
    0x03ee: 'href',
    0x03ef: 'rel',
    0x03f0: 'rev',
    0x03f1: 'type',
    0x03f9: 'media',
    0x03fa: 'target',
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x978e: 'onerror',
    0x9791: 'onload',
    }
ATTRS61 = {
    0x9399: 'clear',
    }
ATTRS62 = {
    0x8001: 'name',
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS63 = {
    0x1771: 'scrolldelay',
    0x1772: 'direction',
    0x1773: 'behavior',
    0x1774: 'scrollamount',
    0x1775: 'loop',
    0x1776: 'vspace',
    0x1777: 'hspace',
    0x1778: 'truespeed',
    0x8006: 'width',
    0x8007: 'height',
    0x9785: 'onbounce',
    0x978b: 'onfinish',
    0x978c: 'onstart',
    0xfe0c: 'bgcolor',
    }
ATTRS65 = {
    0x03ea: 'http-equiv',
    0x03eb: 'content',
    0x03ec: 'url',
    0x03f6: 'charset',
    0x8001: 'name',
    }
ATTRS66 = {
    0x03f5: 'n',
    }
ATTRS71 = {
    # 0x8000: "border",
    0x8000: 'usemap',
    0x8001: 'name',
    0x8006: 'width',
    0x8007: 'height',
    0x8046: 'title',
    0x804a: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x8bbb: 'classid',
    0x8bbc: 'data',
    0x8bbf: 'codebase',
    0x8bc0: 'codetype',
    0x8bc1: 'code',
    0x8bc2: 'type',
    0x8bc5: 'vspace',
    0x8bc6: 'hspace',
    0x978e: 'onerror',
    }
ATTRS72 = {
    0x03eb: 'compact',
    0x03ec: 'start',
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x939a: 'type',
    }
ATTRS73 = {
    0x03ea: 'selected',
    0x03eb: 'value',
    }
ATTRS74 = {
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x9399: 'clear',
    }
ATTRS75 = {
    # 0x8000: "name",
    # 0x8000: "value",
    0x8000: 'type',
    }
ATTRS76 = {
    0x9399: 'clear',
    }
ATTRS77 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x9399: 'clear',
    }
ATTRS78 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS82 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS83 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS84 = {
    0x03ea: 'src',
    0x03ed: 'for',
    0x03ee: 'event',
    0x03f0: 'defer',
    0x03f2: 'type',
    0x978e: 'onerror',
    }
ATTRS85 = {
    0x03eb: 'size',
    0x03ec: 'multiple',
    0x8000: 'align',
    0x8001: 'name',
    0x978f: 'onchange',
    }
ATTRS86 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS87 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS88 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS89 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS90 = {
    0x03eb: 'type',
    0x03ef: 'media',
    0x8046: 'title',
    0x978e: 'onerror',
    0x9791: 'onload',
    }
ATTRS91 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS92 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS93 = {
    0x03ea: 'cols',
    0x03eb: 'border',
    0x03ec: 'rules',
    0x03ed: 'frame',
    0x03ee: 'cellspacing',
    0x03ef: 'cellpadding',
    0x03fa: 'datapagesize',
    0x8006: 'width',
    0x8007: 'height',
    0x8046: 'title',
    0x804a: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x938a: 'background',
    0x93a5: 'bordercolor',
    0x93a6: 'bordercolorlight',
    0x93a7: 'bordercolordark',
    0xfe0c: 'bgcolor',
    }
ATTRS94 = {
    0x8049: 'align',
    0x93a8: 'valign',
    0xfe0c: 'bgcolor',
    }
ATTRS95 = {
    0x8049: 'align',
    0x93a8: 'valign',
    }
ATTRS96 = {
    0x07d2: 'rowspan',
    0x07d3: 'colspan',
    0x8006: 'width',
    0x8007: 'height',
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x938a: 'background',
    0x938e: 'nowrap',
    0x93a5: 'bordercolor',
    0x93a6: 'bordercolorlight',
    0x93a7: 'bordercolordark',
    0x93a8: 'valign',
    0xfe0c: 'bgcolor',
    }
ATTRS97 = {
    0x1b5a: 'rows',
    0x1b5b: 'cols',
    0x1b5c: 'wrap',
    0x1b5d: 'readonly',
    0x8001: 'name',
    0x977b: 'onselect',
    0x978f: 'onchange',
    }
ATTRS98 = {
    0x8049: 'align',
    0x93a8: 'valign',
    0xfe0c: 'bgcolor',
    }
ATTRS99 = {
    0x07d2: 'rowspan',
    0x07d3: 'colspan',
    0x8006: 'width',
    0x8007: 'height',
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x938a: 'background',
    0x938e: 'nowrap',
    0x93a5: 'bordercolor',
    0x93a6: 'bordercolorlight',
    0x93a7: 'bordercolordark',
    0x93a8: 'valign',
    0xfe0c: 'bgcolor',
    }
ATTRS100 = {
    0x8049: 'align',
    0x93a8: 'valign',
    0xfe0c: 'bgcolor',
    }
ATTRS102 = {
    0x8007: 'height',
    0x8046: 'title',
    0x8049: 'align',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x93a5: 'bordercolor',
    0x93a6: 'bordercolorlight',
    0x93a7: 'bordercolordark',
    0x93a8: 'valign',
    0xfe0c: 'bgcolor',
    }
ATTRS103 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS104 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS105 = {
    0x03eb: 'compact',
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    0x939a: 'type',
    }
ATTRS106 = {
    0x8046: 'title',
    0x804b: 'style',
    0x83ea: 'class',
    0x83eb: 'id',
    }
ATTRS108 = {
    0x9399: 'clear',
    }

TAGS_ATTRS = [
    None,
    None,
    None,
    ATTRS3,   # a
    None,     # acronym
    ATTRS5,   # address
    ATTRS6,   # applet
    ATTRS7,   # area
    ATTRS8,   # b
    ATTRS9,   # base
    ATTRS10,  # basefont
    None,     # bdo
    ATTRS12,  # bgsound
    ATTRS13,  # big
    None,     # blink
    ATTRS15,  # blockquote
    ATTRS16,  # body
    ATTRS17,  # br
    ATTRS18,  # button
    ATTRS19,  # caption
    ATTRS20,  # center
    ATTRS21,  # cite
    ATTRS22,  # code
    ATTRS23,  # col
    ATTRS24,  # colgroup
    None,
    None,
    ATTRS27,  # dd
    None,     # del
    ATTRS29,  # dfn
    None,     # dir
    ATTRS31,  # div
    ATTRS32,  # dl
    ATTRS33,  # dt
    ATTRS34,  # em
    ATTRS35,  # embed
    ATTRS36,  # fieldset
    ATTRS37,  # font
    ATTRS38,  # form
    ATTRS39,  # frame
    ATTRS40,  # frameset
    None,
    ATTRS42,  # h1
    ATTRS43,  # h2
    ATTRS44,  # h3
    ATTRS45,  # h4
    ATTRS46,  # h5
    ATTRS47,  # h6
    None,     # head
    ATTRS49,  # hr
    None,     # html
    ATTRS51,  # i
    ATTRS52,  # iframe
    ATTRS53,  # img
    ATTRS54,  # input
    None,     # ins
    ATTRS56,  # kbd
    ATTRS57,  # label
    ATTRS58,  # legend
    ATTRS59,  # li
    ATTRS60,  # link
    ATTRS61,  # tag61
    ATTRS62,  # map
    ATTRS63,  # tag63
    None,     # tag64
    ATTRS65,  # meta
    ATTRS66,  # nextid
    None,     # nobr
    None,     # noembed
    None,     # noframes
    None,     # noscript
    ATTRS71,  # object
    ATTRS72,  # ol
    ATTRS73,  # option
    ATTRS74,  # p
    ATTRS75,  # param
    ATTRS76,  # plaintext
    ATTRS77,  # pre
    ATTRS78,  # q
    None,     # rp
    None,     # rt
    None,     # ruby
    ATTRS82,  # s
    ATTRS83,  # samp
    ATTRS84,  # script
    ATTRS85,  # select
    ATTRS86,  # small
    ATTRS87,  # span
    ATTRS88,  # strike
    ATTRS89,  # strong
    ATTRS90,  # style
    ATTRS91,  # sub
    ATTRS92,  # sup
    ATTRS93,  # table
    ATTRS94,  # tbody
    ATTRS95,  # tc
    ATTRS96,  # td
    ATTRS97,  # textarea
    ATTRS98,  # tfoot
    ATTRS99,  # th
    ATTRS100,  # thead
    None,     # title
    ATTRS102,  # tr
    ATTRS103,  # tt
    ATTRS104,  # u
    ATTRS105,  # ul
    ATTRS106,  # var
    None,     # wbr
    None,
    ]

MAP = (TAGS, ATTRS0, TAGS_ATTRS)



#!/usr/local/bin/python3.13


__license__   = 'GPL v3'
__copyright__ = '2011, Kovid Goyal <kovid@kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

import json

from qt.core import (
    QAbstractListModel,
    QComboBox,
    QDialog,
    QFormLayout,
    QHBoxLayout,
    QIcon,
    QItemSelectionModel,
    QLineEdit,
    Qt,
    QToolButton,
    QVBoxLayout,
    QWidget,
    pyqtSignal,
)

from calibre.db.constants import NO_SEARCH_LINK
from calibre.ebooks.metadata.book.render import DEFAULT_AUTHOR_LINK
from calibre.ebooks.metadata.search_internet import qquote
from calibre.gui2 import choose_files, choose_save_file, error_dialog
from calibre.gui2.book_details import get_field_list
from calibre.gui2.dialogs.template_dialog import TemplateDialog
from calibre.gui2.preferences import LazyConfigWidgetBase
from calibre.gui2.preferences.coloring import EditRules
from calibre.gui2.ui import get_gui
from calibre.utils.formatter import EvalFormatter


class DefaultAuthorLink(QWidget):

    changed_signal = pyqtSignal()

    def __init__(self, parent):
        QWidget.__init__(self, parent)
        l = QVBoxLayout()
        l.addWidget(self)
        l.setContentsMargins(0, 0, 0, 0)
        l = QFormLayout(self)
        l.setContentsMargins(0, 0, 0, 0)
        l.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        self.choices = c = QComboBox()
        c.setMinimumContentsLength(30)
        for text, data in [
                (_('Search for the author on Goodreads'), 'search-goodreads'),
                (_('Search for the author on Amazon'), 'search-amzn'),
                (_('Search for the author in your calibre library'), 'search-calibre'),
                (_('Search for the author on Wikipedia'), 'search-wikipedia'),
                (_('Search for the author on Google Books'), 'search-google'),
                (_('Search for the book on Goodreads'), 'search-goodreads-book'),
                (_('Search for the book on Amazon'), 'search-amzn-book'),
                (_('Search for the book on Google Books'), 'search-google-book'),
                (_('No author search URL'), NO_SEARCH_LINK),
                (_('Use a custom search URL'), 'url'),
        ]:
            c.addItem(text, data)
        l.addRow(_('Clicking on &author names should:'), c)
        ul = QHBoxLayout()
        self.custom_url = u = QLineEdit(self)
        u.setToolTip(_(
            'Enter the URL to search. It should contain the string {0}'
            '\nwhich will be replaced by the author name. For example,'
            '\n{1}. Note: the author name is already URL-encoded.').format(
                        '{author}', 'https://en.wikipedia.org/w/index.php?search={author}'))
        u.textChanged.connect(self.changed_signal)
        u.setPlaceholderText(_('Enter the URL'))
        ul.addWidget(u)
        u = self.custom_url_button = QToolButton()
        u.setIcon(QIcon.ic('edit_input.png'))
        u.setToolTip(_('Click this button to open the template tester'))
        u.clicked.connect(self.open_template_tester)
        ul.addWidget(u)
        c.currentIndexChanged.connect(self.current_changed)
        l.addRow(ul)
        self.current_changed()
        c.currentIndexChanged.connect(self.changed_signal)

    @property
    def value(self):
        k = self.choices.currentData()
        if k == 'url':
            return self.custom_url.text()
        return k if k != DEFAULT_AUTHOR_LINK else None

    @value.setter
    def value(self, val):
        i = self.choices.findData(val)
        if i < 0:
            i = self.choices.findData('url')
            self.custom_url.setText(val)
        self.choices.setCurrentIndex(i)

    def open_template_tester(self):
        gui = get_gui()
        db = gui.current_db.new_api
        lv = gui.library_view
        rows = lv.selectionModel().selectedRows()
        if not rows:
            vals = [{'author': qquote(_('Author')), 'title': _('Title'), 'author_sort': _('Author sort')}]
        else:
            vals = []
            for row in rows:
                book_id = lv.model().id(row)
                mi = db.new_api.get_proxy_metadata(book_id)
                vals.append({'author': qquote(mi.authors[0]),
                             'title': qquote(mi.title),
                             'author_sort': qquote(mi.author_sort_map.get(mi.authors[0]))})
        d = TemplateDialog(parent=self, text=self.custom_url.text(), mi=vals, formatter=EvalFormatter)
        if d.exec() == QDialog.DialogCode.Accepted:
            self.custom_url.setText(d.rule[1])

    def current_changed(self):
        k = self.choices.currentData()
        self.custom_url.setVisible(k == 'url')
        self.custom_url_button.setVisible(k == 'url')

    def restore_defaults(self):
        self.value = DEFAULT_AUTHOR_LINK


class DisplayedFields(QAbstractListModel):

    def __init__(self, db, parent=None, pref_name=None, category_icons=None):
        self.pref_name = pref_name or 'book_display_fields'
        QAbstractListModel.__init__(self, parent)

        self.fields = []
        self.db = db
        self.changed = False
        self.category_icons = category_icons

    def get_field_list(self, use_defaults=False):
        return get_field_list(self.db.field_metadata, use_defaults=use_defaults, pref_name=self.pref_name)

    def initialize(self, use_defaults=False):
        self.beginResetModel()
        self.fields = [[x[0], x[1]] for x in self.get_field_list(use_defaults=use_defaults)]
        self.endResetModel()
        self.changed = True

    def rowCount(self, *args):
        return len(self.fields)

    def data(self, index, role):
        try:
            field, visible = self.fields[index.row()]
        except:
            return None
        if role == Qt.ItemDataRole.DisplayRole:
            name = field
            try:
                name = self.db.field_metadata[field]['name']
            except:
                pass
            if field == 'path':
                name = _('Folders/path')
            name = field.partition('.')[0][1:] if field.startswith('@') else name
            if not name:
                return field
            return f'{name} ({field})'
        if role == Qt.ItemDataRole.CheckStateRole:
            return Qt.CheckState.Checked if visible else Qt.CheckState.Unchecked
        if role == Qt.ItemDataRole.DecorationRole:
            if self.category_icons:
                icon = self.category_icons.get(field, None)
                if icon is not None:
                    return icon
            if field.startswith('#'):
                return QIcon.ic('column.png')
        return None

    def toggle_all(self, show=True):
        for i in range(self.rowCount()):
            idx = self.index(i)
            if idx.isValid():
                self.setData(idx, Qt.CheckState.Checked if show else Qt.CheckState.Unchecked, Qt.ItemDataRole.CheckStateRole)

    def flags(self, index):
        ans = QAbstractListModel.flags(self, index)
        return ans | Qt.ItemFlag.ItemIsUserCheckable

    def setData(self, index, val, role):
        ret = False
        if role == Qt.ItemDataRole.CheckStateRole:
            self.fields[index.row()][1] = val in (Qt.CheckState.Checked, Qt.CheckState.Checked.value)
            self.changed = True
            ret = True
            self.dataChanged.emit(index, index)
        return ret

    def restore_defaults(self):
        self.initialize(use_defaults=True)

    def commit(self):
        if self.changed:
            self.db.new_api.set_pref(self.pref_name, self.fields)

    def move(self, idx, delta):
        row = idx.row() + delta
        if row >= 0 and row < len(self.fields):
            t = self.fields[row]
            self.fields[row] = self.fields[row-delta]
            self.fields[row-delta] = t
            self.dataChanged.emit(idx, idx)
            idx = self.index(row)
            self.dataChanged.emit(idx, idx)
            self.changed = True
            return idx


class LazyEditRulesBase(LazyConfigWidgetBase):

    rule_set_name = None

    def __init__(self, parent=None):
        super().__init__(parent)
        self.rules_editor = EditRules(parent)
        self.setLayout(self.rules_editor.layout())

    def genesis(self, gui):
        self.gui = gui
        self.rules_editor.changed.connect(self.changed_signal)

    def lazy_initialize(self):
        if not self.rule_set_name:
            raise NotImplementedError('You must define the attribut "rule_set_name" in LazyEditRulesBase subclasses')
        self.load_rule_set(self.rule_set_name)

    def load_rule_set(self, name):
        db = self.gui.current_db
        mi = selected_rows_metadatas()
        self.rules_editor.initialize(db.field_metadata, db.prefs, mi, name)

    def commit(self):
        self.rules_editor.commit(self.gui.current_db.prefs)
        return LazyConfigWidgetBase.commit(self)

    def restore_defaults(self):
        LazyConfigWidgetBase.restore_defaults(self)
        self.rules_editor.clear()
        self.changed_signal.emit()


class ColumnColorRules(LazyEditRulesBase):
    rule_set_name = 'column_color_rules'


class ColumnIconRules(LazyEditRulesBase):
    rule_set_name = 'column_icon_rules'


class GridEmblemnRules(LazyEditRulesBase):
    rule_set_name = 'cover_grid_icon_rules'


def export_layout(in_widget, model=None):
    filename = choose_save_file(in_widget, 'look_feel_prefs_import_export_field_list',
            _('Save column list to file'),
            filters=[(_('Column list'), ['json'])])
    if filename:
        try:
            with open(filename, 'w') as f:
                json.dump(model.fields, f, indent=1)
        except Exception as err:
            error_dialog(in_widget, _('Export field layout'),
                         _('<p>Could not write field list. Error:<br>%s')%err, show=True)


def import_layout(in_widget, model=None):
    filename = choose_files(in_widget, 'look_feel_prefs_import_export_field_list',
            _('Load column list from file'),
            filters=[(_('Column list'), ['json'])])
    if filename:
        try:
            with open(filename[0]) as f:
                fields = json.load(f)
            model.initialize(pref_data_override=fields)
            in_widget.changed_signal.emit()
        except Exception as err:
            error_dialog(in_widget, _('Import layout'),
                         _('<p>Could not read field list. Error:<br>%s')%err, show=True)


def reset_layout(in_widget, model=None):
    model.initialize(use_defaults=True)
    in_widget.changed_signal.emit()


def move_field_up(widget, model):
    idx = widget.currentIndex()
    if idx.isValid():
        idx = model.move(idx, -1)
        if idx is not None:
            sm = widget.selectionModel()
            sm.select(idx, QItemSelectionModel.SelectionFlag.ClearAndSelect)
            widget.setCurrentIndex(idx)


def move_field_down(widget, model):
    idx = widget.currentIndex()
    if idx.isValid():
        idx = model.move(idx, 1)
        if idx is not None:
            sm = widget.selectionModel()
            sm.select(idx, QItemSelectionModel.SelectionFlag.ClearAndSelect)
            widget.setCurrentIndex(idx)


def selected_rows_metadatas():
    rslt = []
    try:
        db = get_gui().current_db
        rows = get_gui().current_view().selectionModel().selectedRows()
        for row in rows:
            if row.isValid():
                rslt.append(db.new_api.get_proxy_metadata(db.data.index_to_id(row.row())))
    except:
        pass
    return rslt



#!/usr/local/bin/python3.13


__license__ = 'GPL v3'
__copyright__ = '2014, Kovid Goyal <kovid at kovidgoyal.net>'

from qt.core import Qt, QTextCursor


def get_text_around_cursor(editor, before=True):
    cursor = editor.textCursor()
    cursor.clearSelection()
    cursor.movePosition((QTextCursor.MoveOperation.StartOfBlock if before else QTextCursor.MoveOperation.EndOfBlock), QTextCursor.MoveMode.KeepAnchor)
    text = editor.selected_text_from_cursor(cursor)
    return cursor, text


get_text_before_cursor = get_text_around_cursor


def get_text_after_cursor(editor):
    return get_text_around_cursor(editor, before=False)


def is_cursor_on_wrapped_line(editor):
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.MoveOperation.StartOfLine)
    sol = cursor.position()
    cursor.movePosition(QTextCursor.MoveOperation.StartOfBlock)
    return sol != cursor.position()


def get_leading_whitespace_on_block(editor, previous=False):
    cursor = editor.textCursor()
    block = cursor.block()
    if previous:
        block = block.previous()
    if block.isValid():
        text = block.text()
        ntext = text.lstrip()
        return text[:len(text)-len(ntext)]
    return ''


def no_modifiers(ev, *args):
    mods = ev.modifiers()
    for mod_mask in args:
        if mods & mod_mask:
            return False
    return True


def test_modifiers(ev, *args):
    mods = ev.modifiers()
    for mod_mask in args:
        if not mods & mod_mask:
            return False
    return True


def smart_home(editor, ev):
    if no_modifiers(ev, Qt.KeyboardModifier.ControlModifier) and not is_cursor_on_wrapped_line(editor):
        cursor, text = get_text_before_cursor(editor)
        cursor = editor.textCursor()
        mode = QTextCursor.MoveMode.KeepAnchor if test_modifiers(ev, Qt.KeyboardModifier.ShiftModifier) else QTextCursor.MoveMode.MoveAnchor
        cursor.movePosition(QTextCursor.MoveOperation.StartOfBlock, mode)
        if text.strip() and text.lstrip() != text:
            # Move to the start of text
            cursor.movePosition(QTextCursor.MoveOperation.NextWord, mode)
        editor.setTextCursor(cursor)
        return True
    return False


def expand_tabs(text, tw):
    return text.replace('\t', ' ' * tw)


def smart_tab_if_whitespace_only_before_cursor(editor, backwards):
    cursor, text = get_text_before_cursor(editor)
    if not text.lstrip():
        # cursor is preceded by only whitespace
        tw = editor.tw
        text = expand_tabs(text, tw)
        if backwards:
            if leading := len(text):
                new_leading = max(0, leading - tw)
                extra = new_leading % tw
                if extra:
                    new_leading += tw - extra
                cursor.insertText(' ' * new_leading)
                return True
        else:
            spclen = len(text) - (len(text) % tw) + tw
            cursor.insertText(' ' * spclen)
            editor.setTextCursor(cursor)
            return True
    return False


def smart_tab_all_blocks_in_selection(editor, backwards):
    cursor = editor.textCursor()
    c = QTextCursor(cursor)
    c.clearSelection()
    c.setPosition(cursor.selectionStart())
    tab_width = editor.tw
    changed = False
    while not c.atEnd() and c.position() <= cursor.selectionEnd():
        c.clearSelection()
        c.movePosition(QTextCursor.MoveOperation.EndOfBlock)
        c.movePosition(QTextCursor.MoveOperation.StartOfBlock, QTextCursor.MoveMode.KeepAnchor)
        c.movePosition(QTextCursor.MoveOperation.StartOfBlock)
        # select leading whitespace
        while not c.atEnd() and c.document().characterAt(c.position()).isspace():
            c.movePosition(QTextCursor.MoveOperation.NextCharacter, QTextCursor.MoveMode.KeepAnchor)
        text = expand_tabs(editor.selected_text_from_cursor(c), tab_width)
        leading = len(text)
        replaced = False
        if backwards:
            if leading:
                new_leading = max(0, leading - tab_width)
                extra = new_leading % tab_width
                if extra:
                    new_leading += tab_width - extra
                replaced = True
        else:
            new_leading = leading + tab_width
            new_leading -= new_leading % tab_width
            replaced = True
        if replaced:
            c.insertText(' ' * new_leading)
            changed = True
        c.movePosition(QTextCursor.MoveOperation.NextBlock)
        c.movePosition(QTextCursor.MoveOperation.StartOfBlock)
    return changed


def smart_tab(editor, ev):
    cursor = editor.textCursor()
    backwards = ev.key() == Qt.Key.Key_Backtab or (ev.key() == Qt.Key.Key_Tab and bool(ev.modifiers() & Qt.KeyboardModifier.ShiftModifier))
    if cursor.hasSelection():
        return smart_tab_all_blocks_in_selection(editor, backwards)
    return smart_tab_if_whitespace_only_before_cursor(editor, backwards)


def smart_backspace(editor, ev):
    if editor.textCursor().hasSelection():
        return False
    cursor, text = get_text_before_cursor(editor)
    if text and not text.lstrip():
        # cursor is preceded by only whitespace
        tw = editor.tw
        text = expand_tabs(text, tw)
        spclen = max(0, len(text) - (len(text) % tw) - tw)
        cursor.insertText(' ' * spclen)
        editor.setTextCursor(cursor)
        return True
    return False



#!/usr/local/bin/python3.13


__license__   = 'GPL v3'
__copyright__ = '2009, Kovid Goyal <kovid@kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

import os
import re

from qt.core import QApplication, QPixmap

from calibre.ebooks.conversion.config import OPTIONS
from calibre.ebooks.metadata import MetaInformation, string_to_authors, title_sort
from calibre.ebooks.metadata.opf2 import metadata_to_opf
from calibre.gui2 import choose_images, error_dialog
from calibre.gui2.convert import Widget
from calibre.gui2.convert.metadata_ui import Ui_Form
from calibre.library.comments import comments_to_html
from calibre.ptempfile import PersistentTemporaryFile
from calibre.utils.config import tweaks
from calibre.utils.icu import sort_key


def create_opf_file(db, book_id, opf_file=None):
    mi = db.get_metadata(book_id, index_is_id=True)
    old_cover = mi.cover
    mi.cover = None
    mi.application_id = mi.uuid
    raw = metadata_to_opf(mi)
    mi.cover = old_cover
    if opf_file is None:
        opf_file = PersistentTemporaryFile('.opf')
    opf_file.write(raw)
    opf_file.close()
    return mi, opf_file


def create_cover_file(db, book_id):
    cover = db.cover(book_id, index_is_id=True)
    cf = None
    if cover:
        cf = PersistentTemporaryFile('.jpeg')
        cf.write(cover)
        cf.close()
    return cf


class MetadataWidget(Widget, Ui_Form):

    TITLE = _('Metadata')
    ICON  = 'dialog_information.png'
    HELP  = _('Set the metadata. The output file will contain as much of this '
            'metadata as possible.')
    COMMIT_NAME = 'metadata'

    def __init__(self, parent, get_option, get_help, db=None, book_id=None):
        Widget.__init__(self, parent, OPTIONS['pipe']['metadata'])
        self.db, self.book_id = db, book_id
        self.cover_changed = False
        self.cover_data = None
        if self.db is not None:
            self.initialize_metadata_options()
        self.initialize_options(get_option, get_help, db, book_id)
        self.cover_button.clicked.connect(self.select_cover)
        self.comment.hide_toolbars()
        self.cover.cover_changed.connect(self.change_cover)
        self.series.currentTextChanged.connect(self.series_changed)
        cuh = self.db.new_api.pref('categories_using_hierarchy', default=())
        if 'series' in cuh:
            self.series.set_hierarchy_separator('.')
        if 'tags' in cuh:
            self.tags.set_hierarchy_separator('.')
        self.cover.draw_border = False

    def change_cover(self, data):
        self.cover_changed = True
        self.cover_data = data

    def deduce_author_sort(self, *args):
        au = str(self.author.currentText())
        au = re.sub(r'\s+et al\.$', '', au)
        authors = string_to_authors(au)
        self.author_sort.setText(self.db.author_sort_from_authors(authors))
        self.author_sort.home(False)

    def initialize_metadata_options(self):
        self.initialize_combos()
        self.author.editTextChanged.connect(self.deduce_author_sort)

        mi = self.db.get_metadata(self.book_id, index_is_id=True)
        self.title.setText(mi.title), self.title.home(False)
        self.publisher.show_initial_value(mi.publisher if mi.publisher else '')
        self.publisher.home(False)
        self.author_sort.setText(mi.author_sort if mi.author_sort else '')
        self.author_sort.home(False)
        self.tags.setText(', '.join(mi.tags if mi.tags else []))
        self.tags.update_items_cache(self.db.new_api.all_field_names('tags'))
        self.tags.home(False)
        self.comment.html = comments_to_html(mi.comments) if mi.comments else ''
        self.series.show_initial_value(mi.series if mi.series else '')
        self.series.home(False)
        if mi.series_index is not None:
            try:
                self.series_index.setValue(mi.series_index)
            except:
                self.series_index.setValue(1.0)

        cover = self.db.cover(self.book_id, index_is_id=True)
        if cover:
            pm = QPixmap()
            pm.loadFromData(cover)
            if not pm.isNull():
                pm.setDevicePixelRatio(self.devicePixelRatio())
                self.cover.setPixmap(pm)
                self.cover_data = cover
                self.set_cover_tooltip(pm)
        else:
            pm = QApplication.instance().cached_qpixmap('default_cover.png', device_pixel_ratio=self.devicePixelRatio())
            self.cover.setPixmap(pm)
            self.cover.setToolTip(_('This book has no cover'))
        for x in ('author', 'series', 'publisher'):
            x = getattr(self, x)
            x.lineEdit().deselect()
        self.series_changed()

    def series_changed(self):
        self.series_index.setEnabled(len(self.series.currentText().strip()) > 0)

    def set_cover_tooltip(self, pm):
        tt = _('Cover size: %(width)d x %(height)d pixels') % dict(
                width=pm.width(), height=pm.height())
        self.cover.setToolTip(tt)

    def initialize_combos(self):
        self.initalize_authors()
        self.initialize_series()
        self.initialize_publisher()

    def initalize_authors(self):
        all_authors = self.db.all_authors()
        all_authors.sort(key=lambda x: sort_key(x[1]))
        self.author.set_separator('&')
        self.author.set_space_before_sep(True)
        self.author.set_add_separator(tweaks['authors_completer_append_separator'])
        self.author.update_items_cache(self.db.new_api.all_field_names('authors'))

        au = self.db.authors(self.book_id, True)
        if not au:
            au = _('Unknown')
        au = ' & '.join([a.strip().replace('|', ',') for a in au.split(',')])
        self.author.show_initial_value(au)
        self.author.home(False)

    def initialize_series(self):
        self.series.set_separator(None)
        self.series.update_items_cache(self.db.new_api.all_field_names('series'))

    def initialize_publisher(self):
        self.publisher.set_separator(None)
        self.publisher.update_items_cache(self.db.new_api.all_field_names('publisher'))

    def get_title_and_authors(self):
        title = str(self.title.text()).strip()
        if not title:
            title = _('Unknown')
        authors = str(self.author.text()).strip()
        authors = string_to_authors(authors) if authors else [_('Unknown')]
        return title, authors

    def get_metadata(self):
        title, authors = self.get_title_and_authors()
        mi = MetaInformation(title, authors)
        publisher = str(self.publisher.text()).strip()
        if publisher:
            mi.publisher = publisher
        author_sort = str(self.author_sort.text()).strip()
        if author_sort:
            mi.author_sort = author_sort
        comments = self.comment.html
        if comments:
            mi.comments = comments
        mi.series_index = float(self.series_index.value())
        series = str(self.series.currentText()).strip()
        if series:
            mi.series = series
        tags = [t.strip() for t in str(self.tags.text()).strip().split(',')]
        if tags:
            mi.tags = tags

        return mi

    def select_cover(self):
        files = choose_images(self, 'change cover dialog',
                             _('Choose cover for ') + str(self.title.text()))
        if not files:
            return
        _file = files[0]
        if _file:
            _file = os.path.abspath(_file)
            if not os.access(_file, os.R_OK):
                d = error_dialog(self.parent(), _('Cannot read'),
                        _('You do not have permission to read the file: ') + _file)
                d.exec()
                return
            cover = None
            try:
                with open(_file, 'rb') as f:
                    cover = f.read()
            except OSError as e:
                d = error_dialog(self.parent(), _('Error reading file'),
                        _('<p>There was an error reading from file: <br /><b>') + _file + '</b></p><br />'+str(e))
                d.exec()
            if cover:
                pix = QPixmap()
                pix.loadFromData(cover)
                pix.setDevicePixelRatio(getattr(self, 'devicePixelRatioF', self.devicePixelRatio)())
                if pix.isNull():
                    d = error_dialog(self.parent(), _('Error reading file'),
                                      _file + _(' is not a valid picture'))
                    d.exec()
                else:
                    self.cover_path.setText(_file)
                    self.set_cover_tooltip(pix)
                    self.cover.setPixmap(pix)
                    self.cover_changed = True
                    self.cpixmap = pix
                    self.cover_data = cover

    def get_recommendations(self):
        return {
            'prefer_metadata_cover': bool(self.opt_prefer_metadata_cover.isChecked()),
        }

    def pre_commit_check(self):
        if self.db is None:
            return True
        db = self.db.new_api
        title, authors = self.get_title_and_authors()
        try:
            if title != db.field_for('title', self.book_id):
                db.set_field('title', {self.book_id:title})
                langs = db.field_for('languages', self.book_id)
                if langs:
                    db.set_field('sort', {self.book_id:title_sort(title, langs[0])})
            if list(authors) != list(db.field_for('authors', self.book_id)):
                db.set_field('authors', {self.book_id:authors})
            if self.cover_changed and self.cover_data is not None:
                self.db.set_cover(self.book_id, self.cover_data)
        except OSError as err:
            err.locking_violation_msg = _("Failed to change on disk location of this book's files.")
            raise
        publisher = self.publisher.text().strip()
        if publisher != db.field_for('publisher', self.book_id):
            db.set_field('publisher', {self.book_id:publisher})
        author_sort = self.author_sort.text().strip()
        if author_sort != db.field_for('author_sort', self.book_id):
            db.set_field('author_sort', {self.book_id:author_sort})
        tags = [t.strip() for t in self.tags.text().strip().split(',')]
        if tags != list(db.field_for('tags', self.book_id)):
            db.set_field('tags', {self.book_id:tags})
        series_index = float(self.series_index.value())
        series = self.series.currentText().strip()
        if series != db.field_for('series', self.book_id):
            db.set_field('series', {self.book_id:series})
        if series and series_index != db.field_for('series_index', self.book_id):
            db.set_field('series_index', {self.book_id:series_index})
        return True

    def commit(self, save_defaults=False):
        '''
        Settings are stored in two attributes: `opf_file` and `cover_file`.
        Both may be None. Also returns a recommendation dictionary.
        '''
        recs = self.commit_options(save_defaults)
        self.user_mi = self.get_metadata()
        self.cover_file = self.opf_file = None
        if self.db is not None:
            self.mi, self.opf_file = create_opf_file(self.db, self.book_id)
            cover = self.db.cover(self.book_id, index_is_id=True)
            if cover:
                cf = PersistentTemporaryFile('.jpeg')
                cf.write(cover)
                cf.close()
                self.cover_file = cf
        return recs



__license__ = 'GPL 3'
__copyright__ = '2010, Hiroshi Miura <miurahr@linux.com>'
__docformat__ = 'restructuredtext en'
__all__ = ['Unihandecoder']

'''
Decode unicode text to an ASCII representation of the text.
Translate unicode characters to ASCII.

Inspired from John Schember's unidecode library which was created as part
of calibre.

Copyright(c) 2009, John Schember

Tranliterate the string from unicode characters to ASCII in Chinese and others.

'''
import unicodedata


class Unihandecoder:
    preferred_encoding = None
    decoder = None

    def __init__(self, lang='zh', encoding='utf-8'):
        self.preferred_encoding = encoding
        lang = lang.lower()
        if lang[:2] == 'ja':
            from calibre.ebooks.unihandecode.jadecoder import Jadecoder
            self.decoder = Jadecoder()
        elif lang[:2] == 'kr' or lang == 'korean':
            from calibre.ebooks.unihandecode.krdecoder import Krdecoder
            self.decoder = Krdecoder()
        elif lang[:2] == 'vn' or lang == 'vietnum':
            from calibre.ebooks.unihandecode.vndecoder import Vndecoder
            self.decoder = Vndecoder()
        else:  # zh and others
            from calibre.ebooks.unihandecode.unidecoder import Unidecoder
            self.decoder = Unidecoder()

    def decode(self, text):
        if isinstance(text, bytes):
            try:
                text = text.decode(self.preferred_encoding)
            except Exception:
                text = text.decode('utf-8', 'replace')
        # at first unicode normalize it. (see Unicode standards)
        ntext = unicodedata.normalize('NFKC', text)
        return self.decoder.decode(ntext)



#!/usr/local/bin/python3.13
# License: GPL v3 Copyright: 2019, Kovid Goyal <kovid at kovidgoyal.net>


from qt.core import QKeySequence, QMainWindow, Qt

key_name_to_qt_name = {
    'ArrowRight': 'Right',
    'ArrowLeft': 'Left',
    'ArrowUp': 'Up',
    'ArrowDown': 'Down',
    'PageUp': 'PgUp',
    'PageDown': 'PgDown',
}


def get_main_window_for(widget):
    p = widget
    while p is not None:
        if isinstance(p, QMainWindow):
            return p
        p = p.parent()


def index_to_key_sequence(idx):
    mods = []
    for i, x in enumerate(('ALT', 'CTRL', 'META', 'SHIFT')):
        if idx[i] == 'y':
            mods.append(x.capitalize())
    key = idx[4:]
    mods.append(key_name_to_qt_name.get(key, key))
    return QKeySequence('+'.join(mods))


def key_to_text(key):
    return QKeySequence(key).toString(QKeySequence.SequenceFormat.PortableText).lower()


def ev_to_index(ev):
    m = ev.modifiers()
    mods = []
    for x in (
            Qt.KeyboardModifier.AltModifier, Qt.KeyboardModifier.ControlModifier,
            Qt.KeyboardModifier.MetaModifier, Qt.KeyboardModifier.ShiftModifier):
        mods.append('y' if m & x else 'n')
    return ''.join(mods) + key_to_text(ev.key())


def get_shortcut_for(widget, ev):
    mw = get_main_window_for(widget)
    if mw is None:
        return
    smap = mw.web_view.shortcut_map
    idx = ev_to_index(ev)
    return smap.get(idx)



#!/usr/local/bin/python3.13


__license__ = 'GPL v3'
__copyright__ = '2013, Kovid Goyal <kovid at kovidgoyal.net>'

from collections import namedtuple

from qt.core import (
    QApplication,
    QBrush,
    QCheckBox,
    QColor,
    QColorDialog,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFont,
    QFormLayout,
    QHBoxLayout,
    QIcon,
    QLabel,
    QLineEdit,
    QPalette,
    QPixmap,
    QPushButton,
    QScrollArea,
    QSize,
    QSplitter,
    Qt,
    QTextCharFormat,
    QToolButton,
    QVBoxLayout,
    QWidget,
    pyqtSignal,
)

from calibre.gui2 import error_dialog
from calibre.gui2.tweak_book import tprefs
from calibre.gui2.tweak_book.editor import syntax_text_char_format
from calibre.gui2.tweak_book.widgets import Dialog
from polyglot.builtins import iteritems

underline_styles = {'single', 'dash', 'dot', 'dash_dot', 'dash_dot_dot', 'wave', 'spell'}

_default_theme = None


def default_theme():
    global _default_theme
    if _default_theme is None:
        isdark = QApplication.instance().palette().color(QPalette.ColorRole.WindowText).lightness() > 128
        _default_theme = 'wombat-dark' if isdark else 'pyte-light'
    return _default_theme


# The solarized themes {{{
SLDX = {'base03':'1c1c1c', 'base02':'262626', 'base01':'585858', 'base00':'626262', 'base0':'808080', 'base1':'8a8a8a', 'base2':'e4e4e4', 'base3':'ffffd7', 'yellow':'af8700', 'orange':'d75f00', 'red':'d70000', 'magenta':'af005f', 'violet':'5f5faf', 'blue':'0087ff', 'cyan':'00afaf', 'green':'5f8700'}  # noqa: E501
SLD  = {'base03':'002b36', 'base02':'073642', 'base01':'586e75', 'base00':'657b83', 'base0':'839496', 'base1':'93a1a1', 'base2':'eee8d5', 'base3':'fdf6e3', 'yellow':'b58900', 'orange':'cb4b16', 'red':'dc322f', 'magenta':'d33682', 'violet':'6c71c4', 'blue':'268bd2', 'cyan':'2aa198', 'green':'859900'}  # noqa: E501
m = {f'base{n}':f'base{n:02}' for n in range(1, 4)}
m.update({f'base{n:02}':f'base{n}' for n in range(1, 4)})
SLL =  {m.get(k, k): v for k, v in iteritems(SLD)}
SLLX = {m.get(k, k): v for k, v in iteritems(SLDX)}
SOLARIZED = \
    '''
    CursorLine   bg={base02}
    CursorColumn bg={base02}
    ColorColumn  bg={base02}
    HighlightRegion bg={base00}
    MatchParen   bg={base02} fg={magenta}
    Pmenu        fg={base0} bg={base02}
    PmenuSel     fg={base01} bg={base2}

    Cursor       fg={base03} bg={base0}
    Normal       fg={base0} bg={base02}
    LineNr       fg={base01} bg={base02}
    LineNrC      fg={magenta}
    Visual       fg={base01} bg={base03}

    Comment      fg={base01} italic
    Todo         fg={magenta} bold
    String       fg={cyan}
    Constant     fg={cyan}
    Number       fg={cyan}
    PreProc      fg={orange}
    Identifier   fg={blue}
    Function     fg={blue}
    Type         fg={yellow}
    Statement    fg={green} bold
    Keyword      fg={green}
    Special      fg={red}
    SpecialCharacter bg={base03}

    Error        us=wave uc={red}
    SpellError   us=wave uc={orange}
    Tooltip      fg=black bg=ffffed
    Link         fg={blue}
    BadLink      fg={cyan} us=wave uc={red}

    DiffDelete   bg={base02} fg={red}
    DiffInsert   bg={base02} fg={green}
    DiffReplace  bg={base02} fg={blue}
    DiffReplaceReplace bg={base03}
    '''
# }}}

THEMES = {
    'wombat-dark':  # {{{
    '''
    CursorLine   bg={cursor_loc}
    CursorColumn bg={cursor_loc}
    ColorColumn  bg={cursor_loc}
    HighlightRegion bg=3d3d3d
    MatchParen   bg=444444
    Pmenu        fg=f6f3e8 bg=444444
    PmenuSel     fg=yellow bg={identifier}
    Tooltip      fg=black bg=ffffed

    Cursor       bg=656565
    Normal       fg=f6f3e8 bg=242424
    LineNr       fg=857b6f bg=000000
    LineNrC      fg=yellow
    Visual       fg=black bg=888888

    Comment      fg={comment}
    Todo         fg=8f8f8f
    String       fg={string}
    Constant     fg={constant}
    Number       fg={constant}
    PreProc      fg={constant}
    Identifier   fg={identifier}
    Function     fg={identifier}
    Type         fg={identifier}
    Statement    fg={keyword}
    Keyword      fg={keyword}
    Special      fg={special}
    Error        us=wave uc=red
    SpellError   us=wave uc=orange
    SpecialCharacter bg=666666
    Link         fg=cyan
    BadLink      fg={string} us=wave uc=red

    DiffDelete   bg=341414 fg=642424
    DiffInsert   bg=143414 fg=246424
    DiffReplace  bg=141434 fg=242464
    DiffReplaceReplace bg=002050

    '''.format(
        cursor_loc='323232',
        identifier='cae682',
        comment='99968b',
        string='95e454',
        keyword='8ac6f2',
        constant='e5786d',
        special='e7f6da'),  # }}}

    'pyte-light':  # {{{
    '''
    CursorLine   bg={cursor_loc}
    CursorColumn bg={cursor_loc}
    ColorColumn  bg={cursor_loc}
    HighlightRegion bg=E3F988
    MatchParen   bg=cfcfcf
    Pmenu        fg=white bg=808080
    PmenuSel     fg=white bg=808080
    Tooltip      fg=black bg=ffffed

    Cursor       fg=black bg=b0b4b8
    Normal       fg=404850 bg=f0f0f0
    LineNr       fg=white bg=8090a0
    LineNrC      fg=yellow
    Visual       fg=white bg=8090a0

    Comment      fg={comment} italic
    Todo         fg={comment} italic bold
    String       fg={string}
    Constant     fg={constant}
    Number       fg={constant}
    PreProc      fg={constant}
    Identifier   fg={identifier}
    Function     fg={identifier}
    Type         fg={identifier}
    Statement    fg={keyword}
    Keyword      fg={keyword}
    Special      fg={special} italic
    SpecialCharacter bg=afafaf
    Error        us=wave uc=red
    SpellError   us=wave uc=magenta
    Link         fg=blue
    BadLink      fg={string} us=wave uc=red

    DiffDelete   bg=rgb(255,180,200) fg=rgb(200,80,110)
    DiffInsert   bg=rgb(180,255,180) fg=rgb(80,210,80)
    DiffReplace  bg=rgb(206,226,250) fg=rgb(90,130,180)
    DiffReplaceReplace bg=rgb(180,210,250)

    '''.format(
        cursor_loc='F8DE7E',
        identifier='7b5694',
        comment='a0b0c0',
        string='4070a0',
        keyword='007020',
        constant='a07040',
        special='70a0d0'),  # }}}

    'solarized-x-dark': SOLARIZED.format(**SLDX),
    'solarized-dark': SOLARIZED.format(**SLD),
    'solarized-light': SOLARIZED.format(**SLL),
    'solarized-x-light': SOLARIZED.format(**SLLX),

}


def read_color(col):
    if QColor.isValidColor(col):
        return QBrush(QColor(col))
    if col.startswith('rgb('):
        r, g, b = map(int, (x.strip() for x in col[4:-1].split(',')))
        return QBrush(QColor(r, g, b))
    try:
        r, g, b = col[0:2], col[2:4], col[4:6]
        r, g, b = int(r, 16), int(g, 16), int(b, 16)
        return QBrush(QColor(r, g, b))
    except Exception:
        pass


Highlight = namedtuple('Highlight', 'fg bg bold italic underline underline_color')


def read_theme(raw):
    ans = {}
    for line in raw.splitlines():
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        bold = italic = False
        fg = bg = name = underline = underline_color = None
        line = line.partition('#')[0]
        for i, token in enumerate(line.split()):
            if i == 0:
                name = token
            else:
                if token == 'bold':
                    bold = True
                elif token == 'italic':
                    italic = True
                elif '=' in token:
                    prefix, val = token.partition('=')[0::2]
                    if prefix == 'us':
                        underline = val if val in underline_styles else None
                    elif prefix == 'uc':
                        underline_color = read_color(val)
                    elif prefix == 'fg':
                        fg = read_color(val)
                    elif prefix == 'bg':
                        bg = read_color(val)
        if name is not None:
            ans[name] = Highlight(fg, bg, bold, italic, underline, underline_color)
    return ans


THEMES = {k:read_theme(raw) for k, raw in iteritems(THEMES)}


def u(x):
    x = {'spell':'SpellCheck', 'dash_dot':'DashDot', 'dash_dot_dot':'DashDotDot'}.get(x, x.capitalize())
    if 'Dot' in x:
        return x + 'Line'
    return x + 'Underline'


underline_styles = {x:getattr(QTextCharFormat.UnderlineStyle, u(x)) for x in underline_styles}


def to_highlight(data):
    data = data.copy()
    for c in ('fg', 'bg', 'underline_color'):
        data[c] = read_color(data[c]) if data.get(c, None) is not None else None
    return Highlight(**data)


def read_custom_theme(data):
    dt = THEMES[default_theme()].copy()
    dt.update({k:to_highlight(v) for k, v in iteritems(data)})
    return dt


def get_theme(name):
    try:
        return THEMES[name]
    except KeyError:
        try:
            ans = tprefs['custom_themes'][name]
        except KeyError:
            return THEMES[default_theme()]
        else:
            return read_custom_theme(ans)


def highlight_to_char_format(h):
    ans = syntax_text_char_format()
    if h.bold:
        ans.setFontWeight(QFont.Weight.Bold)
    if h.italic:
        ans.setFontItalic(True)
    if h.fg is not None:
        ans.setForeground(h.fg)
    if h.bg is not None:
        ans.setBackground(h.bg)
    if h.underline:
        ans.setUnderlineStyle(underline_styles[h.underline])
        if h.underline_color is not None:
            ans.setUnderlineColor(h.underline_color.color())
    return ans


def theme_color(theme, name, attr):
    try:
        return getattr(theme[name], attr).color()
    except (KeyError, AttributeError):
        return getattr(THEMES[default_theme()][name], attr).color()


def theme_format(theme, name):
    try:
        h = theme[name]
    except KeyError:
        h = THEMES[default_theme()][name]
    return highlight_to_char_format(h)


def custom_theme_names():
    return tuple(tprefs['custom_themes'])


def builtin_theme_names():
    return tuple(THEMES)


def all_theme_names():
    return builtin_theme_names() + custom_theme_names()


# Custom theme creation/editing {{{

class CreateNewTheme(Dialog):

    def __init__(self, parent=None):
        Dialog.__init__(self, _('Create custom theme'), 'custom-theme-create', parent=parent)

    def setup_ui(self):
        self.l = l = QFormLayout(self)
        self.setLayout(l)

        self._name = n = QLineEdit(self)
        l.addRow(_('&Name of custom theme:'), n)

        self.base = b = QComboBox(self)
        b.addItems(sorted(builtin_theme_names()))
        l.addRow(_('&Builtin theme to base on:'), b)
        idx = b.findText(tprefs['editor_theme'] or default_theme())
        if idx == -1:
            idx = b.findText(default_theme())
        b.setCurrentIndex(idx)

        l.addRow(self.bb)

    @property
    def theme_name(self):
        return str(self._name.text()).strip()

    def accept(self):
        if not self.theme_name:
            return error_dialog(self, _('No name specified'), _(
                'You must specify a name for your theme'), show=True)
        if '*' + self.theme_name in custom_theme_names():
            return error_dialog(self, _('Name already used'), _(
                'A custom theme with the name %s already exists') % self.theme_name, show=True)
        return Dialog.accept(self)


def col_to_string(color):
    return '{:02X}{:02X}{:02X}'.format(*color.getRgb()[:3])


class ColorButton(QPushButton):

    changed = pyqtSignal()

    def __init__(self, data, name, text, parent):
        QPushButton.__init__(self, text, parent)
        self.ic = QPixmap(self.iconSize())
        color = data[name]
        self.data, self.name = data, name
        if color is not None:
            self.current_color = read_color(color).color()
            self.ic.fill(self.current_color)
        else:
            self.ic.fill(Qt.GlobalColor.transparent)
            self.current_color = color
        self.update_tooltip()
        self.setIcon(QIcon(self.ic))
        self.clicked.connect(self.choose_color)

    def clear(self):
        self.current_color = None
        self.update_tooltip()
        self.ic.fill(Qt.GlobalColor.transparent)
        self.setIcon(QIcon(self.ic))
        self.data[self.name] = self.value
        self.changed.emit()

    def choose_color(self):
        col = QColorDialog.getColor(self.current_color or Qt.GlobalColor.black, self, _('Choose color'))
        if col.isValid():
            self.current_color = col
            self.update_tooltip()
            self.ic.fill(col)
            self.setIcon(QIcon(self.ic))
            self.data[self.name] = self.value
            self.changed.emit()

    def update_tooltip(self):
        self.setToolTip(_('Red: {0} Green: {1} Blue: {2}').format(*self.current_color.getRgb()[:3]) if self.current_color else _('No color'))

    @property
    def value(self):
        if self.current_color is None:
            return None
        return col_to_string(self.current_color)


class Bool(QCheckBox):

    changed = pyqtSignal()

    def __init__(self, data, key, text, parent):
        QCheckBox.__init__(self, text, parent)
        self.data, self.key = data, key
        self.setChecked(data.get(key, False))
        self.stateChanged.connect(self._changed)

    def _changed(self, state):
        self.data[self.key] = self.value
        self.changed.emit()

    @property
    def value(self):
        return self.checkState() == Qt.CheckState.Checked


class Property(QWidget):

    changed = pyqtSignal()

    def __init__(self, name, data, parent=None):
        QWidget.__init__(self, parent)
        self.l = l = QHBoxLayout(self)
        self.setLayout(l)
        self.label = QLabel(name)
        l.addWidget(self.label)
        self.data = data

        def create_color_button(key, text):
            b = ColorButton(data, key, text, self)
            b.changed.connect(self.changed), l.addWidget(b)
            bc = QToolButton(self)
            bc.setIcon(QIcon.ic('clear_left.png'))
            bc.setToolTip(_('Remove color'))
            bc.clicked.connect(b.clear)
            h = QHBoxLayout()
            h.addWidget(b), h.addWidget(bc)
            return h

        for k, text in (('fg', _('&Foreground')), ('bg', _('&Background'))):
            h = create_color_button(k, text)
            l.addLayout(h)

        for k, text in (('bold', _('B&old')), ('italic', _('&Italic'))):
            w = Bool(data, k, text, self)
            w.changed.connect(self.changed)
            l.addWidget(w)

        self.underline = us = QComboBox(self)
        us.addItems(sorted(tuple(underline_styles) + ('',)))
        idx = us.findText(data.get('underline', '') or '')
        us.setCurrentIndex(max(idx, 0))
        us.currentIndexChanged.connect(self.us_changed)
        self.la = la = QLabel(_('&Underline:'))
        la.setBuddy(us)
        h = QHBoxLayout()
        h.addWidget(la), h.addWidget(us), l.addLayout(h)

        h = create_color_button('underline_color', _('Color'))
        l.addLayout(h)
        l.addStretch(1)

    def us_changed(self):
        self.data['underline'] = str(self.underline.currentText()) or None
        self.changed.emit()


# Help text {{{

HELP_TEXT = _('''\
<h2>Creating a custom theme</h2>

<p id="attribute" lang="und">You can create a custom syntax highlighting theme, \
with your own colors and font styles. The most important types of highlighting \
rules are described below. Note that not every rule supports every kind of \
customization, for example, changing font or underline styles for the \
<code>Cursor</code> rule does not have any effect as that rule is used only for \
the color of the blinking cursor.</p>

<p>As you make changes to your theme on the left, the changes will be reflected live in this panel.</p>

<p xml:lang="und">
{Normal}
    The most important rule. Sets the foreground and background colors for the \
    editor as well as the style of "normal" text, that is, text that does not match any special syntax.

{Visual}
    Defines the colors for text selected by the mouse.

{CursorLine}
    Defines the color for the line containing the cursor.

{LineNr}
    Defines the colors for the line numbers on the left.

{MatchParen}
    Defines the colors for matching tags in HTML and matching
    braces in CSS.

{Function}
    Used for highlighting tags in HTML

{Type}
    Used for highlighting attributes in HTML

{Statement}
    Tag names in HTML

{Constant}
    Namespace prefixes in XML and constants in CSS

{SpecialCharacter}
    Non{endash}breaking{nbsp}spaces/hyphens in HTML

{Error}
    Syntax errors such as <this <>

{SpellError}
    Misspelled words such as <span lang="en">thisword</span>

{Comment}
    Comments like <!-- this one -->

</p>

<style type="text/css">
/* Some CSS so you can see how the highlighting rules affect it */

p.someclass {{
    font-family: serif;
    font-size: 12px;
    line-height: 1.2;
}}
</style>
''')  # }}}


class ThemeEditor(Dialog):

    def __init__(self, parent=None):
        Dialog.__init__(self, _('Create/edit custom theme'), 'custom-theme-editor', parent=parent)

    def setup_ui(self):
        self.block_show = False
        self.properties = []
        self.l = l  = QVBoxLayout(self)
        self.setLayout(l)
        h = QHBoxLayout()
        l.addLayout(h)
        self.la = la = QLabel(_('&Edit theme:'))
        h.addWidget(la)
        self.theme = t = QComboBox(self)
        la.setBuddy(t)
        t.addItems(sorted(custom_theme_names()))
        t.setMinimumWidth(200)
        if t.count() > 0:
            t.setCurrentIndex(0)
        t.currentIndexChanged.connect(self.show_theme)
        h.addWidget(t)

        self.add_button = b = QPushButton(QIcon.ic('plus.png'), _('Add &new theme'), self)
        b.clicked.connect(self.create_new_theme)
        h.addWidget(b)

        self.remove_button = b = QPushButton(QIcon.ic('minus.png'), _('&Remove theme'), self)
        b.clicked.connect(self.remove_theme)
        h.addWidget(b)
        h.addStretch(1)

        self.scroll = s = QScrollArea(self)
        self.w = w = QWidget(self)
        s.setWidget(w), s.setWidgetResizable(True)
        self.cl = cl = QVBoxLayout()
        w.setLayout(cl)

        from calibre.gui2.tweak_book.editor.text import TextEdit
        self.preview = p = TextEdit(self, expected_geometry=(73, 50))
        t = {x: f'<b>{x}</b>' for x in (
            'Normal', 'Visual', 'CursorLine', 'LineNr', 'MatchParen',
            'Function', 'Type', 'Statement', 'Constant', 'SpecialCharacter',
            'Error', 'SpellError', 'Comment')
        }
        t['nbsp'] = '\xa0'
        t['endash'] = '–'
        p.load_text(HELP_TEXT.format(**t))
        p.setMaximumWidth(p.size_hint.width() + 5)
        s.setMinimumWidth(600)
        self.splitter = sp = QSplitter(self)
        l.addWidget(sp)
        sp.addWidget(s), sp.addWidget(p)

        self.bb.clear()
        self.bb.addButton(QDialogButtonBox.StandardButton.Close)
        l.addWidget(self.bb)

        if self.theme.count() > 0:
            self.show_theme()

    def update_theme(self, name):
        data = tprefs['custom_themes'][name]
        extra = set(data) - set(THEMES[default_theme()])
        missing = set(THEMES[default_theme()]) - set(data)
        for k in extra:
            data.pop(k)
        for k in missing:
            data[k] = dict(THEMES[default_theme()][k]._asdict())
            for nk, nv in iteritems(data[k]):
                if isinstance(nv, QBrush):
                    data[k][nk] = str(nv.color().name())
        if extra or missing:
            tprefs['custom_themes'][name] = data
        return data

    def show_theme(self):
        if self.block_show:
            return
        for c in self.properties:
            c.changed.disconnect()
            self.cl.removeWidget(c)
            c.setParent(None)
            c.deleteLater()
        self.properties = []
        name = str(self.theme.currentText())
        if not name:
            return
        data = self.update_theme(name)
        maxw = 0
        for k in sorted(data):
            w = Property(k, data[k], parent=self)
            w.changed.connect(self.changed)
            self.properties.append(w)
            maxw = max(maxw, w.label.sizeHint().width())
            self.cl.addWidget(w)
        for p in self.properties:
            p.label.setMinimumWidth(maxw), p.label.setMaximumWidth(maxw)
        self.preview.apply_theme(read_custom_theme(data))

    @property
    def theme_name(self):
        return str(self.theme.currentText())

    def changed(self):
        name = self.theme_name
        data = self.update_theme(name)
        self.preview.apply_theme(read_custom_theme(data))

    def create_new_theme(self):
        d = CreateNewTheme(self)
        if d.exec() == QDialog.DialogCode.Accepted:
            name = '*' + d.theme_name
            base = str(d.base.currentText())
            theme = {}
            for key, val in iteritems(THEMES[base]):
                theme[key] = {k:col_to_string(v.color()) if isinstance(v, QBrush) else v for k, v in iteritems(val._asdict())}
            tprefs['custom_themes'][name] = theme
            tprefs['custom_themes'] = tprefs['custom_themes']
            t = self.theme
            self.block_show = True
            t.clear(), t.addItems(sorted(custom_theme_names()))
            t.setCurrentIndex(t.findText(name))
            self.block_show = False
            self.show_theme()

    def remove_theme(self):
        name = self.theme_name
        if name:
            tprefs['custom_themes'].pop(name, None)
            tprefs['custom_themes'] = tprefs['custom_themes']
            t = self.theme
            self.block_show = True
            t.clear(), t.addItems(sorted(custom_theme_names()))
            if t.count() > 0:
                t.setCurrentIndex(0)
            self.block_show = False
            self.show_theme()

    def sizeHint(self):
        g = self.screen().availableSize()
        return QSize(min(1500, g.width() - 25), 650)
# }}}


if __name__ == '__main__':
    from calibre.gui2 import Application
    app = Application([])
    d = ThemeEditor()
    d.exec()
    del app



#!/usr/local/bin/python3.13


__license__   = 'GPL v3'
__copyright__ = '2010, Kovid Goyal <kovid@kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

import os
import shutil
import sys
import time

from calibre import isbytestring
from calibre.constants import filesystem_encoding, islinux, ismacos, iswindows

recycle = None

if iswindows:
    from threading import Lock

    from calibre.utils.ipc import eintr_retry_call
    from calibre_extensions import winutil
    recycler = None
    rlock = Lock()

    def start_recycler():
        global recycler
        if recycler is None:
            from calibre.utils.ipc.simple_worker import start_pipe_worker
            recycler = start_pipe_worker('from calibre.utils.recycle_bin import recycler_main; recycler_main()')

    def recycle_path(path):
        winutil.move_to_trash(str(path))

    def recycler_main():
        stdin = getattr(sys.stdin, 'buffer', sys.stdin)
        stdout = getattr(sys.stdout, 'buffer', sys.stdout)
        while True:
            path = eintr_retry_call(stdin.readline)
            if not path:
                break
            try:
                path = path.decode('utf-8').rstrip()
            except (ValueError, TypeError):
                break
            try:
                recycle_path(path)
            except:
                eintr_retry_call(stdout.write, b'KO\n')
                stdout.flush()
                try:
                    import traceback
                    traceback.print_exc()  # goes to stderr, which is the same as for parent process
                except Exception:
                    pass  # Ignore failures to write the traceback, since GUI processes on windows have no stderr
            else:
                eintr_retry_call(stdout.write, b'OK\n')
                stdout.flush()

    def delegate_recycle(path):
        if '\n' in path:
            raise ValueError(f'Cannot recycle paths that have newlines in them ({path!r})')
        with rlock:
            start_recycler()
            recycler.stdin.write(path.encode('utf-8'))
            recycler.stdin.write(b'\n')
            recycler.stdin.flush()
            # Theoretically this could be made non-blocking using a
            # thread+queue, however the original implementation was blocking,
            # so I am leaving it as blocking.
            result = eintr_retry_call(recycler.stdout.readline)
            if result.rstrip() != b'OK':
                raise RuntimeError(f'recycler failed to recycle: {path!r}')

    def recycle(path):
        # We have to run the delete to recycle bin in a separate process as the
        # morons who wrote SHFileOperation designed it to spin the event loop
        # even when no UI is created. And there is no other way to send files
        # to the recycle bin on windows. Le Sigh. So we do it in a worker
        # process. Unfortunately, if the worker process exits immediately after
        # deleting to recycle bin, winblows does not update the recycle bin
        # icon. Le Double Sigh. So we use a long lived worker process, that is
        # started on first recycle, and sticks around to handle subsequent
        # recycles.
        if isinstance(path, bytes):
            path = path.decode(filesystem_encoding)
        path = os.path.abspath(path)  # Windows does not like recycling relative paths
        return delegate_recycle(path)

elif ismacos:
    from calibre_extensions.cocoa import send2trash

    def osx_recycle(path):
        if isbytestring(path):
            path = path.decode(filesystem_encoding)
        send2trash(path)
    recycle = osx_recycle
elif islinux:
    from calibre.utils.linux_trash import send2trash

    def fdo_recycle(path):
        if isbytestring(path):
            path = path.decode(filesystem_encoding)
        path = os.path.abspath(path)
        send2trash(path)
    recycle = fdo_recycle

can_recycle = callable(recycle)


def nuke_recycle():
    global can_recycle
    can_recycle = False


def restore_recyle():
    global can_recycle
    can_recycle = callable(recycle)


def delete_file(path, permanent=False):
    if not permanent and can_recycle:
        try:
            recycle(path)
            return
        except:
            import traceback
            traceback.print_exc()
    os.remove(path)


def delete_tree(path, permanent=False):
    if permanent:
        try:
            # For completely mysterious reasons, sometimes a file is left open
            # leading to access errors. If we get an exception, wait and hope
            # that whatever has the file (Antivirus, DropBox?) lets go of it.
            shutil.rmtree(path)
        except:
            import traceback
            traceback.print_exc()
            time.sleep(1)
            shutil.rmtree(path)
    else:
        if can_recycle:
            try:
                recycle(path)
                return
            except:
                import traceback
                traceback.print_exc()
        delete_tree(path, permanent=True)



#!/usr/local/bin/python3.13

__license__ = 'GPL v3'
__copyright__ = '2014, Kovid Goyal <kovid at kovidgoyal.net>'


def control(func):
    func.function_type = 'control'
    return func


def data(func):
    func.function_type = 'data'
    return func


class DataError(Exception):

    def __init__(self, tb, msg=None):
        Exception.__init__(self, msg or _('Failed to get completion data'))
        self.tb = tb

    def traceback(self):
        return str(self) + '\n' + self.tb



#!/usr/local/bin/python3.13


__license__   = 'GPL v3'
__copyright__ = '2008, Kovid Goyal kovid@kovidgoyal.net'
__docformat__ = 'restructuredtext en'

from qt.core import QCheckBox, QDialog, QDialogButtonBox, QHBoxLayout, QIcon, QLabel, QMenu, Qt, QVBoxLayout

from calibre import confirm_config_name
from calibre.gui2 import dynamic
from calibre.gui2.dialogs.message_box import Icon


class Dialog(QDialog):

    def __init__(
        self, msg, name, parent, config_set=dynamic, icon='dialog_warning.png',
        title=None, confirm_msg=None, show_cancel_button=True, extra_button=None,
        extra_button_choices=None
    ):
        QDialog.__init__(self, parent)
        self.setWindowTitle(title or _('Are you sure?'))
        self.setWindowIcon(QIcon.ic(icon))
        self.l = l = QVBoxLayout(self)
        self.h = h = QHBoxLayout()
        l.addLayout(h)

        self.icon_widget = Icon(self)
        self.icon_widget.set_icon(QIcon.ic(icon))

        self.msg = m = QLabel(self)
        m.setOpenExternalLinks(True)
        m.setMinimumWidth(350), m.setWordWrap(True), m.setObjectName('msg')
        m.setMaximumHeight(400)
        m.setText(msg)

        h.addWidget(self.icon_widget), h.addSpacing(10), h.addWidget(m)

        self.again = a = QCheckBox((confirm_msg or _('&Show this warning again')), self)
        a.setChecked(True), a.setObjectName('again')
        a.stateChanged.connect(self.toggle)
        l.addWidget(a)

        if show_cancel_button:
            buttons = QDialogButtonBox.StandardButton.Yes | QDialogButtonBox.StandardButton.No
            standard_button = QDialogButtonBox.StandardButton.Yes
        else:
            buttons = QDialogButtonBox.StandardButton.Ok
            standard_button = QDialogButtonBox.StandardButton.Ok
        self.buttonBox = bb = QDialogButtonBox(buttons, self)
        bb.setObjectName('buttonBox')
        bb.setFocus(Qt.FocusReason.OtherFocusReason)
        bb.accepted.connect(self.accept), bb.rejected.connect(self.reject)
        self.extra_button_clicked = False
        self.extra_button_choice = None
        if extra_button:
            b = bb.addButton(extra_button, QDialogButtonBox.ButtonRole.AcceptRole)
            if extra_button_choices:
                m = QMenu()
                b.setMenu(m)
                for acname, text in extra_button_choices.items():
                    ac = m.addAction(text)
                    ac.setObjectName(acname)
                    ac.triggered.connect(self.on_extra_choice_click)
            else:
                b.clicked.connect(self.on_extra_button_click)
        l.addWidget(bb)

        self.name = name
        self.config_set = config_set

        self.resize(self.sizeHint())
        bb.button(standard_button).setFocus(Qt.FocusReason.OtherFocusReason)

    def on_extra_button_click(self):
        self.extra_button_clicked = True

    def on_extra_choice_click(self):
        ac = self.sender()
        self.extra_button_choice = ac.objectName()
        self.accept()

    def toggle(self, *args):
        self.config_set[confirm_config_name(self.name)] = self.again.isChecked()


def confirm(
    msg, name, parent=None, pixmap='dialog_warning.png', title=None,
    show_cancel_button=True, confirm_msg=None, config_set=None, extra_button=None, extra_button_choices=None
):
    config_set = config_set or dynamic
    if not config_set.get(confirm_config_name(name), True):
        if extra_button:
            if extra_button_choices:
                return True, None
            return True, False
        return True
    d = Dialog(msg, name, parent, config_set=config_set, icon=pixmap, extra_button=extra_button,
               title=title, confirm_msg=confirm_msg, show_cancel_button=show_cancel_button, extra_button_choices=extra_button_choices)
    ret = d.exec() == QDialog.DialogCode.Accepted
    if extra_button:
        if extra_button_choices:
            ret = ret, d.extra_button_choice
        else:
            ret = ret, d.extra_button_clicked
    return ret
